from types import SimpleNamespace

import requests

from .config import (
    API_URL,
    DEVELOPER_API_URL,
)


class Response:
    ok: bool
    url: str
    status_code: int

    def __init__(self, response: requests.Response):
        self.response = response

    def __getattr__(self, __name: str):
        return getattr(self.response, __name)

    def __str__(self):
        return f'Response(ok: {self.ok}, status_code: {self.status_code}, url: {self.url} errors: {self.errors})'

    @property
    def errors(self):
        if self.status_code >= 400:
            try:
                return self.json()
            except ValueError as e:
                try:
                    self.response.raise_for_status()
                except Exception as e:
                    return {'error': str(e)}
        return {}


class Session:
    __token = None
    __api_url = None

    def __init__(self, token: str, developer_mode=False):
        self.login(token, developer_mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.__token = None
        self.__api_url = None

    def login(self, token: str, developer_mode=False):
        self.__token = token
        if developer_mode is True:
            self.__api_url = DEVELOPER_API_URL
        else:
            self.__api_url = API_URL

    def logout(self):
        self.__token = None
        self.__api_url = None

    @property
    def is_authenticated(self):
        return self.__token is not None

    @property
    def headers(self):
        return {
            'accept': 'application/json',
            'content-type': 'application/json',
            'authorization': f'Bearer {self.__token}'
        }

    def get(self, url):
        response = requests.get(url, headers=self.headers)
        return Response(response)

    def post(self, url, data=None, json=None):
        response = requests.post(url, data=data, json=json, headers=self.headers)
        return Response(response)

    def patch(self, url, data=None, json=None):
        response = requests.patch(url, data=data, json=json, headers=self.headers)
        return Response(response)

    def build_url(self, endpoint: str, **query: dict):
        query_string = ''
        if query:
            query_string = '?' + '&'.join([f'{k}={v}' for k,v in query.items()])
        return f'{self.__api_url}{endpoint}{query_string}'

    def get_provider_info(self):
        """Obtener la información necesaria para la postulación

        Para iniciar el proceso de certificación, se debe comenzar con la
        postulación y para esto es necesario obtener la información de Alanube
        y la información de las url de recepción.
        https://developer.alanube.co/docs/proceso-de-certificaci%C3%B3n

        Return:
        -------------
        ```json
        {
            "data": {
                "softwareType": {
                    "description": "Tipo de software",
                    "value": "EXTERNO"
                },
                "softwareName": {
                    "description": "Nombre del software",
                    "value": "Alanube"
                },
                "softwareVersion": {
                    "description": "version",
                    "value": 1
                },
                "providerData": {
                    "description": "Datos del proveedor",
                    "rnc": {
                        "description": "RNC/Cedula del proveedor",
                        "value": 132109122
                    },
                    "name": {
                        "description": "Nombre/Razón social del proveedor",
                        "value": "Alanube Soluciones SRL"
                    },
                    "tradeName": {
                        "description": "Nombre comercial",
                        "value": "Alanube Soluciones"
                    }
                }
            }
        }
        ```
        """
        url = self.build_url('provider-info')
        res = self.get(url)
        return res.json()

    def sign_document(self, id_company):
        """Firmado de XML de postulación.
        https://developer.alanube.co/docs/proceso-de-certificaci%C3%B3n

        Return:
        ----------
        ```json
        {
            "signedDocumentUrl": "http:/api-alanube-e-provider-dom-prod/..."
        }
        ```

        """
        url = f'https://ap.alanube.co/dom/v1/sign-document/idCompany/{id_company}'
        res = self.post(url)
        return res.json()


class Resource(SimpleNamespace):

    def __init__(self, response: requests.Response):
        data = response.json()
        super().__init__(**data)
        self._meta = SimpleNamespace(**{
            'response': response,
            'url': response.url,
            'data': data,
        })

    def __repr__(self):
        def parse(dic):
            if isinstance(dic, dict):
                return '  ' + '  '.join([f'{k}: {parse(v)}\n' for k,v in dic.items()])
            return dic
        return parse(self.as_dict())

    def save(self):
        data = self.as_dict()
        if data.get('id'):
            return self.patch(self._meta.url, data)
        return self.post(self._meta.url, data)

    def as_dict(self):
        data = self.__dict__.copy()
        data.pop('_meta', None)
        return data



import datetime
from decimal import Decimal
from typing import Any, Dict, List, Tuple, Type
import logging

from .exceptions import ValidationError
from .config import (
    CURRENCIES,
    FIT_TYPES,
    ITEM_BILLING_INDICATOR_1,
    ITEM_BILLING_INDICATOR_2,
    ITEM_BILLING_INDICATOR_3,
    ITEM_BILLING_INDICATOR_4,
    ITEM_BILLING_INDICATORS,
    NCF_E,
    NCF_E_31,
    NCF_E_32,
    NCF_E_34,
    NCF_E_41,
    NCF_E_43,
    NCF_E_44,
    NCF_E_45,
    NCF_E_47,
    NCF_SERIES,
    NUMBER_DEFAULT_MAX_VALUE,
    PAYMENT_ACCOUNT_TYPES,
    PAYMENT_METHODS,
    PAYMENT_TYPE_CREDIT,
    PAYMENT_TYPES,
    UNIT_MEASURES,
    DollarSign,
    PercentSign,
)
from .utils import (
    compare_amounts_with_margin,
    count_ncf_sequence,
    format_phone_number,
    get_province_from_municipality,
    is_ncfs_types_equal,
    split_ncf,
    to_camel_case,
    validate_credit_note_indicator,
    validate_date,
    validate_email,
    validate_fit_type,
    validate_income_type,
    validate_item_billing_indicator,
    validate_item_good_service_indicator,
    validate_modification_code,
    validate_municipality,
    validate_ncf,
    validate_number,
    validate_payment_account_type,
    validate_payment_type,
    validate_percentage,
    validate_phone_number,
    validate_province,
    validate_rnc,
    validate_tax_amount_indicator,
    validate_tax_type,
    validate_unit_measure,
    validate_value_in,
    validate_web_site,
)
from .fields import (
    Field,
    IntField,
    NCFField,
    RNCField,
    StrField,
    DecimalField,
    DateField,
    UnitMeasureField,
    EmailField,
    PhoneField,
    FormField,
    ListField,
    ListFormField,
    Attr,
    NOT_IMPLEMENTED,
)


class Form:

    is_form = True
    dgii_name = None

    def __init__(self, **kwargs):
        for name, attr in self.__class__.__dict__.items():
            if not isinstance(attr, Field):
                continue

            field: Field = attr

            try:
                value = kwargs[name]
            except KeyError:
                # El campo tiene un valor default.
                if field.default != NOT_IMPLEMENTED:
                    value = field.get_default(self)
                # El campo puede ser nulo.
                elif field.null:
                    value = None
                # No se indicó el campo en los parámetros (y es obligatorio).
                else:
                    raise AttributeError(
                        f'No se indicó un valor para el campo {name} en {self}.'
                    )

            if value == None and not field.null:
                raise AttributeError(f'El campo {name} no puede ser nulo.')

            if value != None:
                value = self.__apply_field_validation(field, name, value)

            field.parent = self
            setattr(self, name, value)

    def __str__(self):
        return self.__class__.__name__.replace('Form', '')

    def __apply_field_validation(self, field: Field, name: str, value: any):
        value = field.validate(
            form=self,
            name=name,
            value=value,
        )
        return value

    @property
    def data(self):
        return self.get_data()

    def get_data(self, allow_null: bool = False, allow_blank: bool = False):
        """
        Obtiene los datos del objeto y los devuelve en forma de diccionario.

        Args:
            * `allow_null` (bool, optional): Permitir valores nulos.
            * `allow_blank` (bool, optional): Permitir string vacios.
        """
        validated_data = self.validate(vars(self))
        data = {}
        for name in validated_data:
            field = getattr(self.__class__, name)
            value = getattr(self, name)

            # Ejecutar validator personalizado.
            if hasattr(self, f'validate_{name}'):
                value = getattr(self, f'validate_{name}')(value, validated_data)

            # Omitir valores nulos si no se permite
            if value is None and allow_null is False:
                continue

            # Omitir valores en blanco si no se permite
            if value == '' and allow_blank is False:
                continue

            value = self.__parse(field, name, value)

            # Convertir el nombre a CamelCase
            name_camel_case = to_camel_case(name)
            data[name_camel_case] = value
        return data

    def validate(self, data: dict):
        return data

    def __parse(self, field: Field, name: str, value: Any):
        # Obtener los datos del objeto Form recursivamente
        if isinstance(value, Form):
            value = value.get_data()

        # Convertir fechas y datetime a formato ISO
        elif isinstance(value, (datetime.date, datetime.datetime)):
            if isinstance(value, datetime.datetime):
                value = value.date()
            value = value.isoformat()

        # Convertir valores Decimal a float (y reducir los decimales)
        elif isinstance(value, (float, Decimal)):
            decimal_places = field.decimal_places if field.decimal_places else 2
            value = round(float(value), decimal_places) # FIXME: La cantidad de decimales puede variar, hay algunos que llevan 4.

        elif isinstance(value, (list, tuple)):
            value = [self.__parse(field, name, v) for v in value]

        return value


class PaymentForm(Form):
    payment_method = IntField(
        'FormaPago',
        min_value=1,
        max_value=8,
        choices=PAYMENT_METHODS,
    )
    payment_amount = DecimalField(
        'MontoPago',
    )


class IdDocForm(Form):
    """
    Identificación del documento

    Args:
    * `company_id` (str): Identificador de la empresa.
    * `id_doc` (IdDocForm): Documento de identificación.
    * `sender` (SenderForm): Formulario del remitente.
    * `buyer` (BuyerForm): Formulario del comprador.
    * `additional_information` (AdditionalInformationForm, opcional): Información adicional.
    * `transport` (TransportForm, opcional): Formulario de transporte.
    * `totals` (TotalsForm): Formulario de totales.
    * `other_currency` (OtherCurrencyForm, opcional): Formulario de otra moneda.
    * `item_details` (List[ItemDetailForm]): Detalles de los elementos.
    * `subtotals` (List[SubtotalForm], opcional): Subtotales.
    * `discounts_or_surcharges` (List[DiscountsOrSurchargeForm], opcional): Descuentos o recargos.
    * `pagination` (List[PaginationForm], opcional): Paginación.
    * `information_reference` (InformationReferenceForm, opcional): Información de referencia.
    * `config` (ConfigForm, opcional): Configuración.
    * `encf` (str):
    * `sequence_due_date` (datetime.date):
    * `deferred_delivery_indicator` (int):
    * `tax_amount_indicator` (int):
    * `income_type` (int):
    * `payment_type` (int):
    * `payment_deadline` (datetime.date):
    * `payment_term` (str):
    * `payment_forms_table` (List[PaymentForm]):
    * `payment_account_type` (str):
    * `payment_account_number` (str):
    * `bank_payment` (str):
    * `date_from` (datetime.date):
    * `date_until` (datetime.date):
    * `total_pages` (int):
    """
    encf: str = NCFField(
        'eNCF',
        null=False,
    )
    sequence_due_date: datetime.date = DateField(
        'FechaVencimientoSecuencia',
        null=False
    )
    deferred_delivery_indicator: int = IntField(
        'IndicadorEnvioDiferido',
        max_value=1
    )
    tax_amount_indicator: int = IntField(
        'IndicadorMontoGravado',
        max_value=1,
        validators=[
            (validate_tax_amount_indicator, [Attr('encf')]),
        ],
    )
    income_type: int = IntField(
        'TipoIngresos',
        null=False,
        min_value=1,
        max_value=6,
        validators=[
            (validate_income_type, [Attr('encf')]),
        ]
    )
    payment_type: int = IntField(
        'TipoPago',
        null=False,
        min_value=1,
        max_value=3,
        choices=PAYMENT_TYPES,
        validators=[
            validate_payment_type,
        ]
    )
    payment_deadline: datetime.date = DateField(
        'FechaLimitePago',
    )
    payment_term = StrField(
        'TerminoPago',
        max_length=15,
    )
    payment_forms_table: List[PaymentForm] = ListFormField(
        'TablaFormasPago',
        form_class=PaymentForm,
        max_length=7,
    )
    payment_account_type: str = StrField(
        'TipoCuentaPago',
        max_length=2,
        choices=PAYMENT_ACCOUNT_TYPES,
        validators=[
            (validate_payment_account_type, [Attr('encf')]),
        ]
    )
    payment_account_number: str = StrField(
        'NumeroCuentaPago',
        max_length=28,
    )
    bank_payment: str = StrField(
        'BancoPago',
        max_length=75,
    )
    date_from: datetime.date = DateField(
        'FechaDesde',
    )
    date_until: datetime.date = DateField(
        'FechaHasta',
    )
    total_pages: int = IntField(
        'TotalPaginas',
        min_value=1,
        max_value=999,
    )

    def validate(self, data: dict):
        data = super().validate(data)

        if self.payment_deadline and self.payment_type != PAYMENT_TYPE_CREDIT:
            raise ValidationError({
                'payment_deadline': f'La fecha límite de pago no está permitida cuando el tipo de pago no es {PAYMENT_TYPE_CREDIT}'
            })

        return data


class SenderForm(Form):
    """
    Emisor de la factura.
    """
    rnc = RNCField(
        'RNCEmisor',
        null=False,
    )
    company_name = StrField(
        'RazonSocialEmisor',
        null=False,
        max_length=150,
    )
    tradename = StrField(
        'NombreComercial',
        max_length=150,
    )
    branch_office = StrField(
        'Sucursal',
        max_length=20,
    )
    address = StrField(
        'DireccionEmisor',
        null=False,
        max_length=100,
    )
    municipality = StrField(
        'Municipio',
        validators=[
            (validate_municipality, [Attr('province')]),
        ],
    )
    province = StrField(
        'Provincia',
        validators=[
            validate_province,
        ],
    )
    phone_number = ListField(
        'TablaTelefonoEmisor',
        field_class=PhoneField,
        max_length=3,
    )
    mail = EmailField(
        'CorreoEmisor',
    )
    web_site = StrField(
        'WebSite',
        max_length=50,
        validators=[
            validate_web_site,
        ]
    )
    economic_activity = StrField(
        'ActividadEconomica',
        max_length=100,
    )
    seller_code = StrField(
        'CodigoVendedor',
        max_length=60,
    )
    internal_invoice_number = IntField(
        'NumeroFacturaInterna',
        max_value=99999999999999999999,
    )
    internal_order_number = IntField(
        'NumeroPedidoInterno',
        max_value=99999999999999999999,
    )
    sale_area = StrField(
        'ZonaVenta',
        max_length=20,
    )
    sale_route = StrField(
        'RutaVenta',
        max_length=20,
    )
    additional_information_issuer = StrField(
        'InformacionAdicionalEmisor',
        max_length=250,
    )
    stamp_date = DateField(
        'FechaEmision',
        null=False,
    )



class BuyerForm(Form):
    rnc = RNCField(
        'RNCComprador',
    )
    company_name = StrField(
        'RazonSocialComprador',
        null=False,
        max_length=150,
    )
    contact = StrField(
        'ContactoComprador',
        max_length=80,
    )
    mail = EmailField(
        'CorreoComprador',
    )
    address = StrField(
        'DireccionComprador',
        max_length=100,
    )
    municipality = StrField(
        'Municipio',
        validators=[
            (validate_municipality, [Attr('province')]),
        ],
    )
    province = StrField(
        'Provincia',
        validators=[
            validate_province,
        ],
    )
    deliver_date = DateField(
        'FechaEntrega',
    )
    contact_delivery = StrField(
        'ContactoEntrega',
        max_length=100,
    )
    delivery_address = StrField(
        'DireccionEntrega',
        max_length=100,
    )
    additional_phone = PhoneField(
        'TelefonoAdicional',
    )
    purchase_order_date = DateField(
        'FechaOrdenCompra',
    )
    purchase_order_number = StrField(
        'NumeroOrdenCompra',
        max_length=20,
    )
    internal_code = StrField(
        'CodigoInternoComprador',
        max_length=20,
    )
    responsible_for_payment = StrField(
        'ResponsablePago',
        max_length=20,
    )
    additional_information = StrField(
        'InformacionAdicionalComprador',
        max_length=150,
    )


class AdditionalInformationForm(Form):
    """
    Información adicional del documento.
    """
    shipping_date = DateField(
        'FechaEmbarque',
    )
    shipment_number = StrField(
        'NumeroEmbarque',
        max_length=25,
    )
    container_number = StrField(
        'NumeroContenedor',
        max_length=100,
    )
    reference_number = IntField(
        'NumeroReferencia',
    )
    gross_weight = DecimalField(
        'PesoBruto',
    )
    net_weight = DecimalField(
        'PesoNeto',
    )
    gross_weight_unit = UnitMeasureField(
        'UnidadPesoBruto',
    )
    unit_net_weight = UnitMeasureField(
        'UnidadPesoNeto',
    )
    bulk_quantity = DecimalField(
        'CantidadBulto',
    )
    bulk_unit = UnitMeasureField(
        'UnidadBulto',
    )
    bulk_volume = DecimalField(
        'VolumenBulto',
    )
    unit_volume = UnitMeasureField(
        'UnidadVolumen',
    )


class TransportForm(Form):
    """
    Transporte.
    """
    driver = StrField(
        'Conductor',
        max_length=20,
    )
    transport_document = IntField(
        'DocumentoTransporte',
    )
    file = StrField(
        'Ficha',
        max_length=20,
    )
    license_plate = StrField(
        'Placa',
        max_length=7,
    )
    transportation_route = StrField(
        'RutaTransporte',
        max_length=20,
    )
    transportation_zone = StrField(
        'ZonaTransporte',
        max_length=20,
    )
    albaran_number = IntField(
        'NumeroAlbaran',
        max_value=20,
    )


class TotalsAdditionalTaxForm(Form):
    tax_type = IntField(
        'TipoImpuesto',
        null=True,
    )
    additional_tax_rate = DecimalField(
        'TasaImpuestoAdicional',
        max_value=999.99,
    )
    selective_tax_amount_specific_consumption = DecimalField(
        'MontoImpuestoSelectivoConsumoEspecifico',
    )
    amount_selective_consumption_tax_advalorem = DecimalField(
        'MontoImpuestoSelectivoConsumoAdvalorem',
    )
    other_additional_taxes = DecimalField(
        'OtrosImpuestosAdicionales',
    )


class TotalsForm(Form):
    total_taxed_amount: Decimal = DecimalField(
        'MontoGravadoTotal',
    )
    i1_amount_taxed: Decimal = DecimalField(
        'MontoGravadoI1',
        help_text='Total de la suma de valores de Ítems gravados asignados a '
        'ITBIS tasa 1 (tasa 18%), menos descuentos más recargos. 12 Condicional '
        'a que en la línea de detalle exista algún ítem gravado a tasa ITBIS1.'
    )
    i2_amount_taxed: Decimal = DecimalField(
        'MontoGravadoI2',
    )
    i3_amount_taxed: Decimal = DecimalField(
        'MontoGravadoI3',
    )
    exempt_amount: Decimal = DecimalField(
        'MontoExento',
    )
    itbis_s1: int = IntField(
        'ITBIS1',
        max_value=99,
        default=ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_1],
    )
    itbis_s2: int = IntField(
        'ITBIS2',
        max_value=99,
        default=ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_2],
    )
    itbis_s3: int = IntField(
        'ITBIS3',
        max_value=99,
        default=ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_3],
    )
    itbis_total: Decimal = DecimalField(
        'TotalITBIS',
    )
    itbis1_total: Decimal = DecimalField(
        'TotalITBIS1',
    )
    itbis2_total: Decimal = DecimalField(
        'TotalITBIS2',
    )
    itbis3_total: Decimal = DecimalField(
        'TotalITBIS3',
    )
    additional_tax_amount: Decimal = DecimalField(
        'MontoImpuestoAdicional',
    )
    additional_taxes: List[TotalsAdditionalTaxForm] = ListFormField(
        'ImpuestosAdicionales',
        form_class=TotalsAdditionalTaxForm,
        max_length=20,
    )
    total_amount: Decimal = DecimalField(
        'MontoTotal',
    )
    non_billable_amount: Decimal = DecimalField(
        'MontoNoFacturable',
    )
    amount_period: Decimal = DecimalField(
        'MontoPeriodo',
    )
    previous_balance: Decimal = DecimalField(
        'SaldoAnterior',
    )
    amount_advance_payment: Decimal = DecimalField(
        'MontoAvancePago',
    )
    pay_value: Decimal = DecimalField(
        'ValorPagar',
    )
    itbis_total_retained: Decimal = DecimalField(
        'TotalITBISRetenido',
    )
    isr_total_retention: Decimal = DecimalField(
        'TotalISRRetencion',
    )
    itbis_total_perception: Decimal = DecimalField(
        'TotalITBISPercepcion',
    )
    isr_total_perception: Decimal = DecimalField(
        'TotalISRPercepcion',
    )

    def validate_total_amount(self, total_amount: Decimal, data: dict):
        total_amount = Decimal(total_amount)
        total_taxed_amount = self.total_taxed_amount or 0
        exempt_amount = self.exempt_amount or 0
        itbis_total = self.itbis_total or 0
        additional_tax_amount = self.additional_tax_amount or 0
        total_amount_sum = sum((total_taxed_amount, exempt_amount, itbis_total, additional_tax_amount))

        if not compare_amounts_with_margin(total_amount, total_amount_sum):
            raise ValidationError(
                f'El valor de {total_amount=} no es igual a la sumatoria de '
                f'{total_taxed_amount=}, {exempt_amount=}, {itbis_total=} y {additional_tax_amount=}.'
            )

        return total_amount

    def validate_total_taxed_amount(self, total_taxed_amount: Decimal | None, data: dict):
        if total_taxed_amount == None:
            return None
        i1_amount_taxed = self.i1_amount_taxed or 0
        i2_amount_taxed = self.i2_amount_taxed or 0
        i3_amount_taxed = self.i3_amount_taxed or 0
        itbis123_totals = sum((i1_amount_taxed, i2_amount_taxed, i3_amount_taxed))

        if total_taxed_amount != None and total_taxed_amount != itbis123_totals:
            raise ValidationError(
                f'La sumatoria de los valores {i1_amount_taxed=} + {i2_amount_taxed=} + {i3_amount_taxed=} '
                f'no coinciden con el valor de {total_taxed_amount=}.'
            )
        return total_taxed_amount

    def validate_itbis_total(self, itbis_total: Decimal | None, data: dict):
        if not itbis_total:
            return None
        itbis_total = Decimal(itbis_total)

        # Total de la suma de valores de ITBIS a diferentes tasas. Condicional a
        # que exista Total ITBIS Tasa 1, y/o Total ITBIS Tasa 2 y/o Total ITBIS Tasa 3.
        itbis1_total = self.itbis1_total or 0
        itbis2_total = self.itbis2_total or 0
        itbis3_total = self.itbis3_total or 0

        if itbis_total != sum((itbis1_total, itbis2_total, itbis3_total)):
            raise ValidationError(
                f'El valor de {itbis_total=} no es igual a la sumatoria de '
                f'{itbis1_total=}, {itbis2_total=} y {itbis3_total=}.'
            )
        return itbis_total


class AdditionalTaxOtherCurrencyForm(Form):
    tax_type_other_currency = IntField(
        'TipoImpuestoOtraMoneda',
        null=False,
    )
    additional_tax_rate_other_currency = DecimalField(
        'TasaImpuestoAdicionalOtraMoneda',
        max_value=999.99,
    )
    selective_tax_amount_specific_consumption_other_currency = DecimalField(
        'MontoImpuestoSelectivoConsumoEspecificoOtraMoneda',
    )
    amount_selective_consumption_tax_advalorem_other_currency = DecimalField(
        'MontoImpuestoSelectivoConsumoAdvaloremOtraMoneda',
    )
    other_additional_taxes_other_currency = DecimalField(
        'OtrosImpuestosAdicionalesOtraMoneda',
    )


class OtherCurrencyForm(Form):
    currency_type = StrField(
        'TipoMoneda',
        null=False,
        max_length=3,
        choices=CURRENCIES,
    )
    exchange_rate = DecimalField(
        'TipoCambio',
        null=False,
        max_value=999.9999,
    )
    total_taxed_amount_other_currency = DecimalField(
        'MontoGravadoTotalOtraMoneda',
    )
    amount_taxed1_other_currency = DecimalField(
        'MontoGravado1OtraMoneda',
    )
    amount_taxed2_other_currency = DecimalField(
        'MontoGravado2OtraMoneda',
    )
    amount_taxed3_other_currency = DecimalField(
        'MontoGravado3OtraMoneda',
    )
    exempt_amount_other_currency = DecimalField(
        'MontoExentoOtraMoneda',
    )
    itbis_total_other_currency = DecimalField(
        'TotalITBISOtraMoneda',
    )
    itbis1_total_other_currency = DecimalField(
        'TotalITBIS1OtraMoneda',
    )
    itbis2_total_other_currency = DecimalField(
        'TotalITBIS2OtraMoneda',
    )
    itbis3_total_other_currency = DecimalField(
        'TotalITBIS3OtraMoneda',
    )
    additional_tax_amount_other_currency = DecimalField(
        'MontoImpuestoAdicionalOtraMoneda',
    )
    additional_taxes_other_currency = ListFormField(
        'ImpuestosAdicionalesOtraMoneda',
        form_class=AdditionalTaxOtherCurrencyForm,
        max_length=20,
    )
    total_amount_other_currency = DecimalField(
        'MontoTotalOtraMoneda',
        null=False,
    )


class OtherCurrencyDetailForm(Form):
    price_other_currency = DecimalField(
        'PrecioOtraMoneda',
        null=False,
    )
    discount_other_currency = DecimalField(
        'DescuentoOtraMoneda',
    )
    surcharge_another_currency = DecimalField(
        'RecargoOtraMoneda',
    )
    amount_item_other_currency = DecimalField(
        'MontoItemOtraMoneda',
        null=False,
    )


class ItemCodeForm(Form):
    code_type = StrField(
        'TipoCodigo',
        max_length=14,
    )
    item_code = StrField(
        'CodigoItem',
        max_length=35,
    )


class RetentionForm(Form):
    indicator_agent_withholding_perception = IntField(
        'IndicadorAgenteRetencionoPercepcion',
        min_value=1,
        max_value=2,
    )
    itbis_amount_withheld = DecimalField(
        'MontoITBISRetenido',
    )
    isr_amount_withheld = DecimalField(
        'MontoISRRetenido',
    )


class SubquantityForm(Form):
    subquantity = DecimalField(
        'Subcantidad',
        max_digits=19,
        decimal_places=3,
    )
    code_subquantity = UnitMeasureField(
        'CodigoSubcantidad',
    )


class SubDiscountForm(Form):
    sub_discount_rate = StrField(
        'TipoSubDescuento',
        max_length=1,
        choices=['%', '$'],
    )
    sub_discount_percentage = DecimalField(
        'SubDescuentoPorcentaje',
        max_value=999.99,
    )
    sub_discount_amount = DecimalField(
        'MontoSubDescuento',
    )


class SubSurchargeForm(Form):
    sub_surcharge_type = StrField(
        'TipoSubRecargo',
        max_length=1,
        choices=['%', '$'],
    )
    sub_surcharge_percentage = DecimalField(
        'SubRecargoPorcentaje',
        max_value=999.99,
    )
    sub_surcharge_amount = DecimalField(
        'MontoSubRecargo',
    )


class TaxForm(Form):
    tax_type = IntField(
        'TipoImpuesto',
        max_value=999,
        validators=[
            validate_tax_type,
        ]
    )
    additional_tax_rate = DecimalField(
        'TasaImpuestoAdicional',
        max_value=999.99,
    )
    selective_tax_amount_specific_consumption = DecimalField(
        'MontoImpuestoSelectivoConsumoEspecifico',
    )
    amount_selective_consumption_tax_advalorem = DecimalField(
        'MontoImpuestoSelectivoConsumoAdvalorem',
    )
    other_additional_taxes = DecimalField(
        'OtrosImpuestosAdicionales',
    )


class ItemDetailForm(Form):
    line_number: int = IntField(
        'NumeroLinea',
        null=False,
    )
    item_code_table: List[ItemCodeForm] = ListFormField(
        'TablaCodigosItem',
        form_class=ItemCodeForm,
        max_length=5,
    )
    billing_indicator: int = IntField(
        'IndicadorFacturacion',
        null=False,
        max_value=4,
        validators=[
            validate_item_billing_indicator,
        ]
    )
    retention: RetentionForm = FormField(
        'Retencion',
        form_class=RetentionForm,
    )
    item_name: str = StrField(
        'NombreItem',
        null=False,
        max_length=80,
    )
    good_service_indicator: int = IntField(
        'IndicadorBienoServicio',
        null=False,
        min_value=1,
        max_value=2,
        validators=[
            validate_item_good_service_indicator,
        ]
    )
    item_description: str = StrField(
        'DescripcionItem',
        max_length=1000,
    )
    quantity_item: Decimal = DecimalField(
        'CantidadItem',
        null=False,
    )
    unit_measure: int = UnitMeasureField(
        'UnidadMedida',
    )
    quantity_reference: Decimal = DecimalField(
        'CantidadReferencia',
    )
    reference_unit: int = UnitMeasureField(
        'UnidadReferencia',
    )
    subquantity_table: List[SubquantityForm] = ListFormField(
        'TablaSubcantidad',
        form_class=SubquantityForm,
        max_length=5,
    )
    degrees_alcohol: Decimal = DecimalField(
        'GradosAlcohol',
        max_value=999.99,
    )
    unit_price_reference: Decimal = DecimalField(
        'PrecioUnitarioReferencia',
    )
    elaboration_date: datetime.date = DateField(
        'FechaElaboracion',
    )
    expiration_date_item: datetime.date = DateField(
        'FechaVencimientoItem',
    )
    unit_price_item: Decimal = DecimalField(
        'PrecioUnitarioItem',
        null=False,
        max_digits=20,
        decimal_places=4,
    )
    discount_amount: Decimal = DecimalField(
        'DescuentoMonto',
    )
    sub_discounts: List[SubDiscountForm] = ListFormField(
        'TablaSubDescuento',
        form_class=SubDiscountForm,
        max_length=12,
    )
    surcharge_amount: Decimal = DecimalField(
        'RecargoMonto',
    )
    sub_surcharge: List[SubSurchargeForm] = ListFormField(
        'TablaSubRecargo',
        form_class=SubSurchargeForm,
        max_length=12,
    )
    additional_taxes: List[TaxForm] = ListFormField(
        'TablaImpuestoAdicional',
        form_class=TaxForm,
        max_length=2,
    )
    other_currency_detail: OtherCurrencyDetailForm = FormField(
        'TablaSubcantidad',
        form_class=OtherCurrencyDetailForm,
    )
    item_amount: Decimal = DecimalField(
        'MontoItem',
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.item_amount = self.get_item_amount()

    def get_item_amount(self) -> Decimal:
        """Obtiene (Precio Unitario * Cantidad) - Monto Descuento + Monto Recargo."""
        return (
            self.get_item_amount_without_discount() -
            (
                (self.discount_amount or 0) +
                (self.surcharge_amount or 0)
            )
        )

    def get_item_amount_without_discount(self) -> Decimal:
        """Obtiene (Precio Unitario * Cantidad)."""
        return self.unit_price_item * self.quantity_item



class SubtotalForm(Form):
    sub_total_number = IntField(
        'NumeroSubTotal',
        min_value=1,
        max_value=20, # Porque solo se pueden incluir 20 repeticiones en la tabla <Subtotales>
    )
    subtotal_description = StrField(
        'DescripcionSubtotal',
        max_length=40,
    )
    order = IntField(
        'Order',
        max_value=99,
    )
    sub_total_amount_taxed_total = DecimalField(
        'SubTotalMontoGravadoTotal',
    )
    sub_total_amount_taxed_i1 = DecimalField(
        'SubTotalMontoGravadoI1'
    )
    sub_total_amount_taxed_i2 = DecimalField(
        'SubTotalMontoGravadoI2',
    )
    sub_total_amount_taxed_i3 = DecimalField(
        'SubTotalMontoGravadoI3',
    )
    # Campo es calculado.
    itbis_sub_total = DecimalField(
        'SubTotaITBIS',
        editable=False,
    )
    itbis_1_sub_total = DecimalField(
        'SubTotaITBIS1',
    )
    itbis_2_sub_total = DecimalField(
        'SubTotaITBIS2',
    )
    itbis_3_sub_total = DecimalField(
        'SubTotaITBIS3',
    )
    sub_total_additional_tax = DecimalField(
        'SubTotalImpuestoAdicional',
    )
    sub_total_exempt = DecimalField(
        'SubTotalExento',
    )
    # Campo calculado.
    sub_total_amount = DecimalField(
        'MontoSubTotal',
        editable=False,
    )
    lines = IntField(
        'Lineas',
        max_value=99,
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.itbis_sub_total = sum((
            self.itbis_1_sub_total or 0,
            self.itbis_2_sub_total or 0,
            self.itbis_3_sub_total or 0
        ))
        self.sub_total_amount = sum((
            self.sub_total_amount_taxed_total or 0,
            self.itbis_sub_total or 0,
            self.sub_total_additional_tax or 0,
            self.sub_total_exempt or 0
        ))


class DiscountsOrSurchargeForm(Form):
    """
    Descuento o recargo.
    """
    line_number = IntField(
        'NumeroLinea',
        min_value=1,
        max_value=20,
    )
    fit_type = StrField(
        'TipoAjuste',
        min_length=1,
        max_length=1,
        choices=FIT_TYPES,
        validators=[
            validate_fit_type,
        ]
    )
    norma_1007_indicator = IntField(
        'IndicadorNorma1007',
        max_value=1,
    )
    description_discount_or_surcharge = StrField(
        'DescripcionDescuentooRecargo',
        max_length=45,
    )
    type_value = StrField(
        'TipoValor',
        max_length=1,
        choices=['%', '$'],
    )
    discount_value_or_surcharge = DecimalField(
        'ValorDescuentooRecargo',
        max_value=999.99,
    )
    discount_amount_or_surcharge = DecimalField(
        'MontoDescuentooRecargo',
    )
    discount_amount_or_surcharge_other_currency = DecimalField(
        'MontoDescuentooRecargoOtraMoneda',
    )
    indicator_billing_discount_or_surcharge = IntField(
        'IndicadorFacturacionDescuentooRecargo',
        max_value=4,
        validators=[
            validate_item_billing_indicator
        ],
    )


class SubtotalAdditionalTaxForm(Form):
    subtotal_selective_tax_for_specific_consumption_page = DecimalField(
        'SubtotalImpuestoSelectivoConsumoEspecificoPagina',
    )
    subtotal_other_tax = DecimalField(
        'SubtotalOtrosImpuesto',
    )


class PaginationForm(Form):
    """
    En esta sección se indica la cantidad de páginas del e-CF en la
    Representación Impresa y cuales ítems estarán en cada una. Cada objeto
    deberá repetirse para el total de páginas especificadas. La sección
    paginación es condicional a que el e-CF contenga más de una página, es
    decir, si la RI contiene una sola página, no debe enviarse.

    Campo DGII <Paginacion>
    """
    page_no = IntField(
        'PaginaNo',
        min_value=1,
        max_value=100,
    )
    no_line_from = IntField(
        'NoLineaDesde',
        min_value=1,
        max_value=999,
    )
    no_line_until = IntField(
        'NoLineaHasta',
        min_value=1,
        max_value=999,
    )
    subtotal_amount_taxed_page = DecimalField('SubtotalMontoGravadoPagina')
    subtotal_amount_taxed_1_page = DecimalField('SubtotalMontoGravado1Pagina')
    subtotal_amount_taxed_2_page = DecimalField('SubtotalMontoGravado2Pagina')
    subtotal_amount_taxed_3_page = DecimalField('SubtotalMontoGravado3Pagina')
    exempt_subtotal_page = DecimalField('SubtotalExentoPagina')
    itbis_subtotal_page = DecimalField('SubtotalItbisPagina')
    itbis_1_subtotal_page = DecimalField('SubtotalItbis1Pagina')
    itbis_2_subtotal_page = DecimalField('SubtotalItbis2Pagina')
    itbis_3_subtotal_page = DecimalField('SubtotalItbis3Pagina')
    subtotal_additional_tax_page = DecimalField('SubtotalImpuestoAdicionalPagina')
    subtotal_additional_tax = FormField(
        'SubtotalImpuestoAdicional',
        form_class=SubtotalAdditionalTaxForm,
    )
    subtotal_amount_page = DecimalField('MontoSubtotalPagina')
    subtotal_non_billable_amount_page = DecimalField('SubtotalMontoNoFacturablePagina')

    def validate(self, data: dict):
        no_line_from = data.get('no_line_from')
        no_line_until = data.get('no_line_until')
        if no_line_from is not None and no_line_until is not None:
            if no_line_from > no_line_until:
                raise ValidationError('El valor de no_line_from debe ser igual o menor a no_line_until.')
        return data


class InformationReferenceForm(Form):
    """
    Información de referencia.
    <InformacionReferencia>
    """
    ncf_modified = NCFField(
        'NCFModificado',
    )
    rnc_other_taxpayer = RNCField(
        'RNCOtroContribuyente',
    )
    ncf_modified_date = DateField(
        'FechaNCFModificado',
    )
    modification_code = IntField(
        'CodigoModificacion',
        min_value=1,
        max_value=5,
        validators=[
            validate_modification_code,
        ]
    )
    reason_for_modification = StrField(
        'RazonModificacion',
        max_length=90,
    )


class PdfForm(Form):
    type = StrField(max_length=7, default='generic')
    note = StrField()


class ConfigForm(Form):
    """
    Configuración adicional del documento electrónico
    """
    pdf = FormField(dgii_name=None, form_class=PdfForm, default=PdfForm())



class InvoiceForm(Form):
    """
    Factura Electrónica.

    Args:
    * `company_id` (str): Identificador de la empresa.
    * `id_doc` (IdDocForm): Documento de identificación.
    * `sender` (SenderForm): Formulario del remitente.
    * `buyer` (BuyerForm): Formulario del comprador.
    * `additional_information` (AdditionalInformationForm, opcional): Información adicional.
    * `transport` (TransportForm, opcional): Formulario de transporte.
    * `totals` (TotalsForm): Formulario de totales.
    * `other_currency` (OtherCurrencyForm, opcional): Formulario de otra moneda.
    * `item_details` (List[ItemDetailForm]): Detalles de los elementos.
    * `subtotals` (List[SubtotalForm], opcional): Subtotales.
    * `discounts_or_surcharges` (List[DiscountsOrSurchargeForm], opcional): Descuentos o recargos.
    * `pagination` (List[PaginationForm], opcional): Paginación.
    * `information_reference` (InformationReferenceForm, opcional): Información de referencia.
    * `config` (ConfigForm, opcional): Configuración.
    """
    company_id: str = StrField(
        null=False,
    )
    id_doc: IdDocForm = FormField(
        'IdDoc',
        null=False,
        form_class=IdDocForm,
    )
    sender: SenderForm = FormField(
        'Emisor',
        null=False,
        form_class=SenderForm,
    )
    buyer: BuyerForm = FormField(
        'Comprador',
        null=False,
        form_class=BuyerForm,
    )
    additional_information: AdditionalInformationForm = FormField(
        'InformacionesAdicionales',
        form_class=AdditionalInformationForm,
    )
    transport: TransportForm = FormField(
        'Transporte',
        form_class=TransportForm,
    )
    totals: TotalsForm = FormField(
        'Totales',
        null=False,
        form_class=TotalsForm,
    )
    other_currency: OtherCurrencyForm = FormField(
        'OtraMoneda',
        form_class=OtherCurrencyForm,
    )
    item_details: List[ItemDetailForm] = ListFormField(
        'DetallesItem',
        null=False,
        form_class=ItemDetailForm,
        max_length=100,
    )
    subtotals: List[SubtotalForm] = ListFormField(
        'Subtotales',
        form_class=SubtotalForm,
        max_length=20,
    )
    discounts_or_surcharges: List[DiscountsOrSurchargeForm] = ListFormField(
        'DescuentosORecargos',
        form_class=DiscountsOrSurchargeForm,
        max_length=20,
    )
    pagination: List[PaginationForm] = ListFormField(
        'Paginacion',
        form_class=PaginationForm,
        max_length=100,
    )
    information_reference: InformationReferenceForm = FormField(
        'InformacionReferencia',
        form_class=InformationReferenceForm,
    )
    config = FormField(
        dgii_name=None,
        form_class=ConfigForm,
        default=ConfigForm(),
    )

    def validate(self, data):
        data = super().validate(data)

        ncf_serie, ncf_code, ncf_sequence = split_ncf(self.id_doc.encf)

        # El RNC del comprador es opcional para algunos tipos de NCF.
        if not self.buyer.rnc and ncf_code in (NCF_E_31, NCF_E_41, NCF_E_45):
            raise ValidationError(
                'El RNC del comprobador es obligatorio para los tipos de '
                'comprobantes 31, 41 y 45.'
            )

        # Validar los totals.
        # Estos comprobantes no aplican para itbis.
        if split_ncf(data['id_doc'].encf)[1] in (NCF_E_43, NCF_E_44, NCF_E_47):
            if (
                self.totals.total_taxed_amount or
                self.totals.i1_amount_taxed or
                self.totals.i2_amount_taxed or
                self.totals.i3_amount_taxed or
                self.totals.itbis_total or
                self.totals.itbis1_total or
                self.totals.itbis2_total or
                self.totals.itbis3_total or
                self.totals.total_taxed_amount
            ):
                raise ValidationError(
                    f'Los tipos de NCF {NCF_E_43}, {NCF_E_44} y {NCF_E_47} no aplican para itbis.'
                )

        # Validar que coincidan los montos con los totales calculados.

        # Total de la suma de valores de Ítems gravados asignados a ITBIS tasa 1
        # (tasa 18%), menos descuentos más recargos. 12 Condicional a que en la
        # línea de detalle exista algún ítem gravado a tasa ITBIS1.
        if self.totals.i1_amount_taxed:
            if not compare_amounts_with_margin(self.totals.i1_amount_taxed, self.items_i1_amount):
                raise ValidationError(
                    f'El campo {self.totals.i1_amount_taxed=} no coincide con '
                    'la sumatoria de los valores del detalle gravado con tasa '
                    f'{ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_1]} '
                    f'menos descuentos mas recargos {self.items_i1_amount=}'
                )
        # Total de la suma de valores de Ítems gravados asignados a ITBIS tasa 2
        # (tasa 16%), menos descuentos más recargos. Condicional a que en la
        # línea de detalle exista algún ítem gravado a tasa ITBIS2.
        if self.totals.i2_amount_taxed:
            if not compare_amounts_with_margin(self.totals.i2_amount_taxed, self.items_i2_amount):
                raise ValidationError(
                    f'El campo {self.totals.i2_amount_taxed=} no coincide con '
                    'la sumatoria de los valores del detalle gravado con tasa '
                    f'{ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_2]} '
                    'menos descuentos mas recargos.'
                )
        # Total de la suma de valores de Ítems gravados asignados a ITBIS tasa 3
        # (tasa 0%), menos descuentos más recargos. Condicional a que en la
        # línea de detalle exista algún ítem gravado a tasa ITBIS3.
        if self.totals.i3_amount_taxed:
            if not compare_amounts_with_margin(self.totals.i3_amount_taxed, self.items_i3_amount):
                raise ValidationError(
                    f'El campo {self.totals.i3_amount_taxed=} no coincide con '
                    'la sumatoria de los valores del detalle gravado con tasa '
                    f'{ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_3]} '
                    'menos descuentos mas recargos.'
                )
        # Total de la suma de valores de ítems exentos, menos descuentos más
        # recargos. Condicional a que en la línea de detalle exista algún ítem exento.
        if self.totals.exempt_amount:
            if not compare_amounts_with_margin(self.totals.exempt_amount, self.items_exempt_amount):
                raise ValidationError(
                    f'El campo {self.totals.exempt_amount=} no coincide con '
                    'la sumatoria de los valores del detalle gravado con tasa '
                    f'{ITEM_BILLING_INDICATORS[ITEM_BILLING_INDICATOR_4]} '
                    'menos descuentos mas recargos.'
                )

        return data

    @property
    def items_amount(self):
        """Sumatoria de `item_amount` para cada item en `item_details`."""
        return sum([(e.item_amount or 0) for e in self.item_details])

    @property
    def items_i1_amount(self):
        """Monto total gravado con `ITEM_BILLING_INDICATOR_1`"""
        return sum([
            (e.item_amount or 0) for e in self.item_details
            if e.billing_indicator == ITEM_BILLING_INDICATOR_1
        ])

    @property
    def items_i2_amount(self):
        """Monto total gravado con `ITEM_BILLING_INDICATOR_2`"""
        return sum([
            (e.item_amount or 0) for e in self.item_details
            if e.billing_indicator == ITEM_BILLING_INDICATOR_2
        ])

    @property
    def items_i3_amount(self):
        """Monto total gravado con `ITEM_BILLING_INDICATOR_3`"""
        return sum([
            (e.item_amount or 0) for e in self.item_details
            if e.billing_indicator == ITEM_BILLING_INDICATOR_3
        ])

    @property
    def items_exempt_amount(self):
        """Monto total gravado con `ITEM_BILLING_INDICATOR_4` (excento)"""
        return sum([
            (e.item_amount or 0) for e in self.item_details
            if e.billing_indicator == ITEM_BILLING_INDICATOR_4
        ])



class CreditNoteIdDocForm(IdDocForm):
    """
    idDoc Form para CreditNoteForm.

    Args:
    * `company_id` (str): Identificador de la empresa.
    * `id_doc` (IdDocForm): Documento de identificación.
    * `sender` (SenderForm): Formulario del remitente.
    * `buyer` (BuyerForm): Formulario del comprador.
    * `additional_information` (AdditionalInformationForm, opcional): Información adicional.
    * `transport` (TransportForm, opcional): Formulario de transporte.
    * `totals` (TotalsForm): Formulario de totales.
    * `other_currency` (OtherCurrencyForm, opcional): Formulario de otra moneda.
    * `item_details` (List[ItemDetailForm]): Detalles de los elementos.
    * `subtotals` (List[SubtotalForm], opcional): Subtotales.
    * `discounts_or_surcharges` (List[DiscountsOrSurchargeForm], opcional): Descuentos o recargos.
    * `pagination` (List[PaginationForm], opcional): Paginación.
    * `information_reference` (InformationReferenceForm, opcional): Información de referencia.
    * `config` (ConfigForm, opcional): Configuración.
    * `encf` (str):
    * `deferred_delivery_indicator` (int):
    * `tax_amount_indicator` (int):
    * `income_type` (int):
    * `payment_type` (int):
    * `payment_deadline` (datetime.date):
    * `date_from` (datetime.date):
    * `date_until` (datetime.date):
    * `total_pages` (int):
    * `credit_note_indicator` (int):
    """
    sequence_due_date = None
    payment_term = None
    payment_forms_table = None
    payment_account_type = None
    payment_account_number = None
    bank_payment = None
    credit_note_indicator: int = IntField(
        'IndicadorNotaCredito',
        null=False,
        min_value=0,
        max_value=1,
        help_text=(
            'Sólo para Notas de Crédito que no tienen derecho a rebajar ITBIS. '
            'El indicador tomará valor 1 si la fecha es mayor a 30 días '
            'calendario de la emisión del e-CF afectado.'
        )
    )

    def validate(self, data):
        data = super().validate(data)

        # FIXME: Pendiente de validación esto:
        # Sólo para Notas de Crédito que no tienen derecho a rebajar
        # ITBIS.El indicador tomará valor 1 si la fecha es mayor a 30 días
        # calendario de la emisión del e-CF afectado.
        # https://dgii.gov.do/cicloContribuyente/facturacion/comprobantesFiscalesElectronicosE-CF/Documentacin%20sobre%20eCF/Formatos%20XML/Formato%20Comprobante%20Fiscal%20Electr%C3%B3nico%20(e-CF)%20v1.0.pdf
        logging.warning('Pendiente validar credit_note_indicator.')

        return data


class CreditNoteForm(InvoiceForm):
    """
    Nota de Crédito Electrónica (34).

    https://developer.alanube.co/v1.0-DOM/reference/createcreditnotes
    """
    id_doc: CreditNoteIdDocForm = FormField(
        'IdDoc',
        null=False,
        form_class=CreditNoteIdDocForm,
    )


class CancellationHeaderForm(Form):
    rnc_sender = RNCField(
        'RncEmisor',
        null=False,
    )
    cancelled_encf_quantity = IntField(
        'CantidadeNCFAnulados',
        null=False,
        min_value=1,
        max_value=9999999999,
    )


class CancellationItemRangeCancelledEnfcForm(Form):
    """
    Tabla que contiene un rango de las secuencias anuladas de manera consecutiva,
    de acuerdo con el tipo de e-CF. Se pueden incluir hasta 10,000 repeticiones.
    """
    dgii_name = 'TablaRangoSecuenciasAnuladaseNCF'

    encf_from = NCFField(
        'SecuenciaeNCFDesde',
        null=False,
        help_text=(
            'Se refiere al Número de Comprobante Fiscal Electrónico (e-NCF) '
            'con el secuencial que inicia el rango de secuencias que será anulado.'
        )
    )
    encf_until = NCFField(
        'SecuenciaeNCFHasta',
        null=False,
        help_text=(
            'Se refiere al Número de Comprobante Fiscal Electrónico (e-NCF) con '
            'el secuencial que finaliza el rango de secuencias que será anulado.'
        )
    )


class CancellationItemForm(Form):
    """
    Anulación se pueden incluir 10 hasta repeticiones.
    """
    dgii_name = 'Anulacion'

    line_number = IntField(
        'NoLinea',
        null=False,
        min_value=1,
        max_value=10,
        help_text='Número de la línea o secuencial. Desde 1 hasta 10 repeticiones.',
    )
    ecf_type = IntField(
        'TipoeCF',
        null=False,
        min_value=31,
        max_value=47,
        choices=[int(code) for code in NCF_SERIES[NCF_E]['codes']],
        help_text='Tipo Comprobante Fiscal Electrónico (e-CF).',
    )
    range_cancelled_enfc = ListFormField(
        'TablaRangoSecuenciasAnuladaseNCF',
        null=False,
        form_class=CancellationItemRangeCancelledEnfcForm,
        max_length=10000,
        help_text=(
            'Tabla que contiene un rango de las secuencias anuladas de manera '
            'consecutiva, de acuerdo con el tipo de e-CF. Se pueden incluir '
            'hasta 10,000 repeticiones.'
        )
    )
    cancelled_encf_quantity = IntField(
        'CantidadeNCFAnulados',
        null=False,
        min_value=1,
        max_value=9999999999,
        help_text=(
            'Cantidad de secuencias de e- NCF que se está anulando. En este '
            'campo se deberá sumar las secuencias colocados en la tabla de '
            'rangos de secuencias anuladas de e-NCF'
        )
    )


class CancellationForm(Form):
    """
    Sirve para emitir anulaciones,
    las cuales se usan para anular rangos de numeración que no se usarán.

    ¿Qué se puede anular?
    ---------------------
        1. Comprobantes que han sido rechazados.

    https://developer.alanube.co/v1.0-DOM/reference/createcancelations
    """

    header = FormField(
        'Encabezado',
        null=False,
        form_class=CancellationHeaderForm,
    )
    cancellations = ListFormField(
        'Anulacion',
        null=False,
        max_length=10,
        form_class=CancellationItemForm,
    )


import unittest
from unittest.mock import MagicMock

from alanube import AlanubeDGII
from alanube.dgii.forms import (
    IdDocForm,
    SenderForm,
    BuyerForm,
    TotalsForm,
    ItemDetailForm,
    ConfigForm,
    InvoiceForm,
    SubDiscountForm,
)


class TestDGII(unittest.TestCase):

    def setUp(self):
        self.alanube = AlanubeDGII('YOUR_TOKEN')

    def test_create_invoice(self):
        id_doc = IdDocForm(
            encf='E310000000005',
            sequence_due_date='2023-01-01',
            tax_amount_indicator=1,
            income_type=1,
            payment_type=1,
            payment_deadline='2023-01-30',
        )

        sender = SenderForm(
            rnc=101123456,
            company_name='My Company SRL',
            tradename='My Company',
            branch_office='Principal',
            address='Calle A, No. 1245, Santo Domingo',
            phone_number='8091234567',
            mail='mycompany@example.com',
            web_site='https://mycompany.example.com',
            seller_code='01',
            internal_invoice_number='123456789',
            stamp_date='2023-01-01',
        )

        buyer = BuyerForm(
            rnc=101123457,
            company_name='His Company SRL',
            contact='José, Tel: 8091234567',
            mail='jose@example.com',
            address='Calle B, 1236, San Cristobal',
        )

        totals = TotalsForm(
            total_amount=1180.00,
            total_taxed_amount=1000,
            amount_period=1000.00,
            non_billable_amount=0,
            i1_amount_taxed=1000.00, # Monto gravado a ITBIS tasa1 (18%).
            i2_amount_taxed=0, # Monto gravado a ITBIS tasa2 (16%).
            i3_amount_taxed=0, # Monto gravado a ITBIS tasa3 (0%).
            exempt_amount=0, # Monto exento.
            itbis_total=180.00, # Total ITBIS
            itbis1_total=180.00, # ITBIS (18%)
            itbis2_total=0, # ITBIS (16%)
            itbis3_total=0, # ITBIS (0%)
            additional_tax_amount=None,
            additional_taxes=None,
            itbis_total_retained=None,
            itbis_total_perception=None,
            isr_total_retention=None,
            isr_total_perception=None,
        )

        items = [
            ItemDetailForm(
                line_number=1,
                item_code_table=None,
                billing_indicator=1,
                retention=None,
                item_name='Café Entero Paquete',
                item_description=None,
                good_service_indicator=1,
                quantity_item=10,
                unit_measure=None,
                subquantity_table=None,
                unit_price_item=100.00,
                discount_amount=0,
                sub_discounts=[
                    SubDiscountForm(
                        sub_discount_rate='%',
                        sub_discount_percentage=10,
                    ),
                    SubDiscountForm(
                        sub_discount_rate='$',
                        sub_discount_percentage=5,
                    )
                ],
                item_amount=1000.00,
            ),
        ]

        config = ConfigForm()

        invoice_form = InvoiceForm(
            company_id='MY_ALANUBE_COMPANY_ID',
            id_doc=id_doc,
            sender=sender,
            buyer=buyer,
            totals=totals,
            item_details=items,
            subtotals=None,
            discounts_or_surcharges=None,
            pagination=None,
            config=config,
        )
        self.alanube.create_invoice(invoice_form)


if __name__ == '__main__':
    unittest.main()





import datetime
from decimal import Decimal
import logging
import re
from typing import Dict, List, Tuple, Iterable

from .exceptions import ValidationError
from .config import (
    CREDIT_NOTE_INDICATOR_DAYS,
    CREDIT_NOTE_INDICATOR_EXPIRED,
    CREDIT_NOTE_INDICATOR_NOT_EXPIRED,
    FIT_TYPES,
    INCOME_TYPES,
    ITEM_GOOD_SERVICE_INDICATORS,
    ITEM_BILLING_INDICATORS,
    MODIFICATION_CODES,
    NCF_B,
    NCF_E,
    NCF_E_32,
    NCF_E_34,
    NCF_E_41,
    NCF_E_43,
    NCF_E_46,
    NCF_E_47,
    NCF_SERIES,
    NUMBER_DEFAULT_MAX_VALUE,
    PAYMENT_ACCOUNT_TYPES,
    PAYMENT_METHOD_BONUS,
    PAYMENT_METHODS,
    PROVINCES_AND_MUNICIPALITIES,
    UNIT_MEASURES,
)


def compare_amounts_with_margin(amount1: Decimal, amount2: Decimal, margin: Decimal = 1):
    dif = abs(amount1 - amount2)
    return dif <= margin


def format_phone_number(number: str):
    """Formatea un número de teléfono en el formato ###-###-####.

    Args:
        number (str): Número de teléfono a formatear.

    Returns:
        str: Número de teléfono formateado en el formato ###-###-####.

    Example:
        >>> phone_number = "8095554444"
        >>> formatted_phone_number = format_phone_number(phone_number)
        >>> print(formatted_phone_number)
        809-555-4444
    """
   # Eliminar todos los caracteres no numéricos
    digits = re.sub(r'\D', '', number)

    # Verificar si la cadena resultante contiene 10 dígitos
    if len(digits) != 10:
        raise ValueError("El número de teléfono debe contener 10 dígitos numéricos.")

    # Formatear la cadena en el formato deseado
    formatted_number = f"{digits[:3]}-{digits[3:6]}-{digits[6:]}"

    return formatted_number


def validate_phone_number(number: str):
    return format_phone_number(number)


def get_credit_note_indicator(date: datetime.date):
    date_more_days = date + datetime.timedelta(days=CREDIT_NOTE_INDICATOR_DAYS)
    if date_more_days < datetime.date.today():
        return CREDIT_NOTE_INDICATOR_EXPIRED
    return CREDIT_NOTE_INDICATOR_NOT_EXPIRED


def validate_credit_note_indicator(indicator: int):
    if not indicator in (CREDIT_NOTE_INDICATOR_EXPIRED, CREDIT_NOTE_INDICATOR_NOT_EXPIRED):
        raise ValidationError('El indicador %s no es válido.' % indicator)
    return indicator


def strdate_to_date(str_date: str, _format='%Y-%m-%d'):
    return datetime.strptime(str_date, _format).date()


def validate_date(date: datetime.date | str):
    if isinstance(date, str):
        date = strdate_to_date(date)
    if isinstance(date, datetime.datetime):
        date = date.date()
    return date


def get_ncf_range(ncf_from: str, ncf_until: str):
    serie1, code1, sequence1 = split_ncf(ncf_from, validate=True)
    serie2, code2, sequence2 = split_ncf(ncf_until, validate=True)

    # Validamos haciendo uso de esta función:
    count_ncf_sequence(ncf_from, ncf_until)

    return [f'{serie1}{code1}{n}' for n in range(int(sequence1), int(sequence2) + 1)]


def count_ncf_sequence(ncf_from: str, ncf_until: str):
    """Cuenta la cantidad de NCFs que hay en un rango de NCFs (secuencia).

    >>> count_ncf_sequence('E310000000001', 'E310000000010')
    10
    >>> count_ncf_sequence('E310000000010', 'E310000000020')
    11
    """
    serie1, code1, sequence1 = split_ncf(ncf_from, validate=True)
    serie2, code2, sequence2 = split_ncf(ncf_until, validate=True)

    sequence1 = int(sequence1)
    sequence2 = int(sequence2)

    if (serie1 != serie2) or (code1 != code2):
        raise ValidationError('Los tipos de NCFs no son iguales.')
    if sequence1 > sequence2:
        raise ValidationError('El NCF inicial es mayor al NCF final.')
    return (sequence2 - sequence1) + 1


def is_ncfs_types_equal(ncf_from: str, ncf_until: str):
    serie1, code1, sequence1 = split_ncf(ncf_from, validate=True)
    serie2, code2, sequence2 = split_ncf(ncf_until, validate=True)

    if (serie1 != serie2) or (code1 != code2):
        return False
    return True


def split_ncf(ncf: str, validate=False):
    """Divide el `ncf` en sus 3 partes (serie, código, sequence).

    Args:
    ----------
    * `ncf` (str): Comprobante fiscal a dividir.
    * `validate` (bool): Si es True, se validará con `validate_ncf(ncf)`.

    Use:
    ----------
    >>> split_ncf('E310000224062')
    ('E', '31', '0000224062')

    """
    if validate is True:
        ncf = validate_ncf(ncf)
    serie = ncf[0]
    code = ncf[1:3]
    sequence = ncf[3:]
    return (serie, code, sequence)


def to_camel_case(name: str):
    words = name.split('_')
    capitalized_words = [word.capitalize() for word in words]
    camel_case_string = ''.join(capitalized_words)
    camel_case_string = camel_case_string[0].lower() + camel_case_string[1:]
    return camel_case_string


def validate_ncf(ncf: str, specific_serie: str = None):
    """Valida que el formato del `ncf` sea el correcto.

    Args:
    -------
    * `specific_serie` (str): (`NCF_B` o `NCF_E`), si se indica, se validará
        también que el `ncf` corresponda con dicha serie.
    """
    assert specific_serie in (None, NCF_B, NCF_E), f'El argumento {specific_serie=} no es válido.'
    serie, code, sequence = split_ncf(ncf, validate=False)

    if specific_serie and specific_serie != serie:
        raise ValidationError('La serie "%s" no es "%s".' % (serie, specific_serie))

    if not sequence.isdigit():
        raise ValidationError('La secuencia "%s" no es numérica.' % sequence)

    if serie == NCF_B:
        if not code in NCF_SERIES[NCF_B]['codes']:
            raise ValidationError('Tipo "%s" no válido para serie "%s".' % (code, serie))
        if len(ncf) != NCF_SERIES[NCF_B]['length']:
            raise ValidationError(
                'Los comprobantes de serie "%s" deben ser longitud %d.'
                % (serie, NCF_SERIES[NCF_B]['length'])
            )
    elif serie == NCF_E:
        if not code in NCF_SERIES[NCF_E]['codes']:
            raise ValidationError('Tipo "%s" no válido para serie "%s".' % (code, serie))
        if len(ncf) != NCF_SERIES[NCF_E]['length']:
            raise ValidationError(
                'Los comprobantes de serie "%s" deben ser longitud %d.'
                % (serie, NCF_SERIES[NCF_E]['length'])
            )
    else:
        raise ValidationError('La serie "%s" no es válida.' % serie)

    return ncf


def validate_value_in(value, _in: Iterable, _default = ...):
    if not value in _in:
        if _default != ...:
            return _default
        raise ValidationError('El valor %s no está dentro del listado.' % value)
    return value


def validate_email(email: str):
    pattern = r'^[\w\.-]+@[\w\.-]+\.\w+$'
    if re.match(pattern, email):
        return email
    raise ValidationError(f'El email {email} no es válido.')


def validate_web_site(web_site):
    pattern = r'^(?:https?:\/\/)?(?:www\.)?([\w\.-]+)(?:\.\w+)+$'
    match = re.match(pattern, web_site)
    if match:
        return match.group(1)
    raise ValidationError(f'El web_site {web_site} no es válido.')


def validate_fit_type(fit_type: str):
    if not fit_type in FIT_TYPES:
        raise ValidationError('El tipo de ajuste %s no es válido' % fit_type)
    return fit_type


def validate_item_good_service_indicator(indicator: int):
    if not indicator in ITEM_GOOD_SERVICE_INDICATORS:
        raise ValidationError('El indicador %s del item no es válido' % indicator)
    return indicator


def validate_item_billing_indicator(indicator: int):
    """Valida que el `indicator` sea uno de estos valores:

        * `0`: No Facturable
        * `1`: ITBIS 1 ítem gravado a ITBIS tasa1 (18%).
        * `2`: ITBIS 2 ítem gravado a ITBIS tasa2 (16%).
        * `3`: ITBIS 3 ítem gravado a ITBIS tasa3 (0%).
        * `4`: Exento (E)
    """
    if not indicator in ITEM_BILLING_INDICATORS:
        raise ValidationError('El indicador %s del item no es válido' % indicator)
    return indicator


def validate_income_type(income_type: int, ncf: str = None):
    """Valida que el tipo de ingreso esté dentro de los permitidos.

    Args:
    --------
        * `income_type` (int, opcional): Si se indica, se comprobará si el tipo
            de ingreso es válido para el tipo de `ncf` indicado.
    """
    # No aplica para estos tipos:
    if ncf and split_ncf(ncf)[1] in (NCF_E_41, NCF_E_43, NCF_E_47):
        logging.warning(f'{income_type=} no aplica para {ncf=}.')
        return None

    if not income_type in INCOME_TYPES:
        raise ValidationError('Tipo de ingreso %s no es válido.' % income_type)

    return income_type


def validate_modification_code(code: int):
    """Valida el código utilizado para indicar si el e-CF del comprobante fiscal
    modificado es con la finalidad de:

    1) Anulación total
    2) Corrige texto del comprobante fiscal modificado
    3) Corrige montos del NCF modificado
    4) Reemplazo NCF emitido en contingencia
    5) Referenciar Factura de Consumo Electrónica.
    """
    if not code in MODIFICATION_CODES:
        raise ValidationError('El código de modificación %s no es válido.' % code)
    return code


def validate_number(
        value: int | float | Decimal | str,
        min_value: Decimal | None = 0,
        max_value: Decimal | None = NUMBER_DEFAULT_MAX_VALUE
    ):
    if isinstance(value, (int, Decimal)):
        pass
    elif isinstance(value, float):
        value = Decimal(str(value))
    elif isinstance(value, str) and value.isdigit():
        value = Decimal(value)
    else:
        raise ValidationError(f'El valor "{value}" no es un valor numérico.')

    if min_value != None and value < min_value:
        raise ValidationError(f'El valor {value} es menor al mínimo {min_value}.')
    if max_value != None and value > max_value:
        raise ValidationError(f'El valor {value} es mayor al máximo {max_value}.')

    return value


def validate_payment_type(payment_type: int):
    if not payment_type in INCOME_TYPES:
        raise ValidationError('Tipo de pago %s no es válido.' % payment_type)
    return payment_type


def validate_payment_method(payment_method: int):
    if not payment_method in INCOME_TYPES:
        raise ValidationError('Forma de pago %s no es válida.' % payment_method)
    return payment_method


def validate_payment_account_type(payment_account_type: str, ncf: str = None):
    if payment_account_type == None:
        return None # El tipo de cuenta es opcional.
    if ncf and payment_account_type and split_ncf(ncf)[1] in (NCF_E_34, NCF_E_43):
        raise ValidationError('El tipo de cuenta debe ser None para los tipos de NCF 34 y 43.')
    if not payment_account_type in PAYMENT_ACCOUNT_TYPES:
        raise ValidationError(
            'El tipo de cuenta de pago %s no es válido.' % payment_account_type
        )
    return payment_account_type


# def validate_payment_form(payment: Dict[str, int | float | Decimal], ncf: str = None):
#     if len(payment) != 2 or not payment.get('paymentMethod') or not payment.get('paymentAmount'):
#         raise ValidationError('Debe tener solo 2 items (`paymentMethod` y `paymentAmount`).')

#     payment_method = validate_payment_method(payment['paymentMethod'])
#     payment_amount = validate_number(payment['paymentAmount'])
#     payment['paymentMethod'] = payment_method
#     payment['paymentAmount'] = payment_amount

#     if ncf:
#         ncf_serie, ncf_code, ncf_sequence = split_ncf(ncf)
#         ncf_not_valid_for_payment_method = (NCF_E_34, NCF_E_43)

#         if ncf_code in ncf_not_valid_for_payment_method and payment:
#             raise ValidationError(
#                 'La forma de pago no aplica para los tipos de ncf %s y %s' %
#                 ncf_not_valid_for_payment_method
#             )

#         # Si la forma de pago corresponde al tipo 5 (Bonos o Certificados de regalo)
#         # el e-CF debe ser tipo 32 (Factura Consumo Electrónica).
#         if payment_method == PAYMENT_METHOD_BONUS and ncf_code != NCF_E_32:
#             raise ValidationError(
#                 'Si la forma de pago corresponde al tipo 5 (Bonos o Certificados '
#                 'de regalo) el e-CF debe ser tipo 32 (Factura Consumo Electrónica).'
#             )

#     return payment


# def validate_payment_forms(payments: List[Dict[str, float]], ):
#     payment_forms_table_errors = []
#     validated_payments = []

#     for payment in payments:
#         try:
#             validated_payments.append(validate_payment_form(payment))
#         except ValidationError as e:
#             payment_forms_table_errors.append(f'Pago 1: {e}')

#     if payment_forms_table_errors:
#         raise ValidationError('\n'.join(payment_forms_table_errors))

#     return validated_payments


def validate_percentage(value: Decimal):
    if value < 0 or value > 99:
        raise ValidationError('El pocentaje debe estar comprendido entre 0 y 99.')
    return value


def get_province_from_municipality(municipality_code: str):
    for province_code in PROVINCES_AND_MUNICIPALITIES:
        municipalities = PROVINCES_AND_MUNICIPALITIES[province_code]['municipalities']
        if municipality_code in municipalities:
            return province_code
    raise KeyError('No existe el municipio con el código "%s".' % municipality_code)


def validate_municipality(municipality_code: str, province_code: str = None):
    if not municipality_code:
        return None
    municipality_code = str(municipality_code)
    try:
        _province_code = get_province_from_municipality(municipality_code)
    except KeyError as e:
        raise ValidationError(e)
    else:
        if province_code is not None and str(province_code) != _province_code:
            raise ValidationError(
                    'El municipio %s no pertenece al a provincia %s.' %
                    (municipality_code, province_code)
                )
    return municipality_code


def validate_province(province_code: str):
    if not province_code:
        return None
    province_code = str(province_code)
    if not province_code in PROVINCES_AND_MUNICIPALITIES:
        raise ValidationError('El código de prinvinvcia "%s" no existe.' % province_code)
    return province_code


def validate_rnc(rnc: int | str):
    rnc = str(rnc)
    rnc = re.sub(r'\D', '', rnc)
    if len(rnc) != 9 and len(rnc) != 11:
        raise ValidationError('El RNC %s no tiene un formato válido.' % rnc)
    return int(rnc)


def validate_tax_type(tax_type: int):
    raise NotImplementedError()


def validate_tax_amount_indicator(indicator: int, ncf: str = None):
    if not indicator:
        return None
    indicator = int(indicator)
    if not indicator in (0, 1):
        raise ValidationError(
            'El indicador %d no es válido, debe ser %d (no incluye impuestos) '
            'y %d (incluye impuestos)' % (indicator, 0, 1)
        )
    if ncf and split_ncf(ncf)[1] in (NCF_E_43, NCF_E_46, NCF_E_47):
        raise ValidationError(
            'Los tipos de NCF 43, 46 y 47 no aplican para indicar si el monto '
            'tiene impuestos incluidos o no.'
        )
    return indicator


def validate_unit_measure(unit_measure: int):
    if not unit_measure in UNIT_MEASURES:
        raise ValidationError('Unidad de médida %s no permitida.' % unit_measure)
    return unit_measure


from typing import NewType


DollarSign = NewType('$', str)
PercentSign = NewType('%', str)


API_URL = 'https://api.alanube.co/dom/v1/'

DEVELOPER_API_URL = 'https://sandbox.alanube.co/dom/v1/'


# Status
REGISTERED = 'REGISTERED' # Se registró el documento
TO_SEND = 'TO_SEND' # Se registró y está listo para ser enviado
FAILED = 'FAILED' # Falló el envió del documento, normalmente pasará porque se está enviando un eNFC que ya se envió
WAITING_RESPONSE = 'WAITING_RESPONSE' # Fue enviado y se encuentra en espera de respuesta de la DGII
TO_NOTIFY = 'TO_NOTIFY' # Obtuvo respuesta de la DGII y se notificará mediante el webhook registrado
FINISHED = 'FINISHED' # Su proceso ha finalizado

STATUS_LIST = [
    REGISTERED,
    TO_SEND,
    FAILED,
    WAITING_RESPONSE,
    TO_NOTIFY,
    FINISHED,
]


# Currencies
CURRENCIES = dict(
    BRL = 'BRL',    # REAL BRASILENO
    CAD = 'CAD',	# DOLAR CANADIENSE
    CHF = 'CHF',    # FRANCO SUIZO
    CHY = 'CHY',	# YUAN CHINO
    XDR = 'XDR',	# DERECHO ESPECIAL DE GIRO83
    DKK = 'DKK',	# CORONA DANESA
    ERU = 'EUR',	# EURO
    GBP = 'GBP',	# LIBRA ESTERLINA
    JPY = 'JPY',	# YEN JAPONES
    NOK = 'NOK',	# CORONA NORUEGA
    SCP = 'SCP',	# LIBRA ESCOCESA
    SEK = 'SEK',	# CORONA SUECA
    USD = 'USD',	# DOLAR ESTADOUNIDENSE
    VEF = 'VEF',	# BOLIVAR FUERTE VENEZOLANO
)


UNIT_MEASURES = {
    1: {'Abrev.': 'BARR', 'Medida': 'Barril'},
    2: {'Abrev.': 'BOL', 'Medida': 'Bolsa'},
    3: {'Abrev.': 'BOT', 'Medida': 'Bote'},
    4: {'Abrev.': 'BULTO', 'Medida': 'Bultos'},
    5: {'Abrev.': 'BOTELLA', 'Medida': 'Botella'},
    6: {'Abrev.': 'CAJ', 'Medida': 'Caja'},
    7: {'Abrev.': 'CAJETILLA', 'Medida': 'Cajetilla'},
    8: {'Abrev.': 'CM', 'Medida': 'Centímetro'},
    9: {'Abrev.': 'CIL', 'Medida': 'Cilindro'},
    10: {'Abrev.': 'CONJ', 'Medida': 'Conjunto'},
    11: {'Abrev.': 'CONT', 'Medida': 'Contenedor'},
    12: {'Abrev.': 'DÍA', 'Medida': 'Día'},
    13: {'Abrev.': 'DOC', 'Medida': 'Docena'},
    14: {'Abrev.': 'FARD', 'Medida': 'Fardo'},
    15: {'Abrev.': 'GL', 'Medida': 'Galones'},
    16: {'Abrev.': 'GRAD', 'Medida': 'Grado'},
    17: {'Abrev.': 'GR', 'Medida': 'Gramo'},
    18: {'Abrev.': 'GRAN', 'Medida': 'Granel'},
    19: {'Abrev.': 'HOR', 'Medida': 'Hora'},
    20: {'Abrev.': 'HUAC', 'Medida': 'Huacal'},
    21: {'Abrev.': 'KG', 'Medida': 'Kilogramo'},
    22: {'Abrev.': 'kWh', 'Medida': 'Kilovatio Hora'},
    23: {'Abrev.': 'LB', 'Medida': 'Libra'},
    24: {'Abrev.': 'LITRO', 'Medida': 'Litro'},
    25: {'Abrev.': 'LOT', 'Medida': 'Lote'},
    26: {'Abrev.': 'M', 'Medida': 'Metro'},
    27: {'Abrev.': 'M2', 'Medida': 'Metro Cuadrado'},
    28: {'Abrev.': 'M3', 'Medida': 'Metro Cúbico'},
    29: {'Abrev.': 'MMBTU', 'Medida': 'Millones de Unidades Térmicas'},
    30: {'Abrev.': 'MIN', 'Medida': 'Minuto'},
    31: {'Abrev.': 'PAQ', 'Medida': 'Paquete'},
    32: {'Abrev.': 'PAR', 'Medida': 'Par'},
    33: {'Abrev.': 'PIE', 'Medida': 'Pie'},
    34: {'Abrev.': 'PZA', 'Medida': 'Pieza'},
    35: {'Abrev.': 'ROL', 'Medida': 'Rollo'},
    36: {'Abrev.': 'SOBR', 'Medida': 'Sobre'},
    37: {'Abrev.': 'SEG', 'Medida': 'Segundo'},
    38: {'Abrev.': 'TANQUE', 'Medida': 'Tanque'},
    39: {'Abrev.': 'TONE', 'Medida': 'Tonelada'},
    40: {'Abrev.': 'TUB', 'Medida': 'Tubo'},
    41: {'Abrev.': 'YD', 'Medida': 'Yarda'},
    42: {'Abrev.': 'YD2', 'Medida': 'Yarda Cuadrada'},
    43: {'Abrev.': 'UND', 'Medida': 'Unidad'},
    44: {'Abrev.': 'EA', 'Medida': 'Elemento'},
    45: {'Abrev.': 'MILLAR', 'Medida': 'Millar'},
    46: {'Abrev.': 'SAC', 'Medida': 'Saco'},
    47: {'Abrev.': 'LAT', 'Medida': 'Lata'},
    48: {'Abrev.': 'DIS', 'Medida': 'Display'},
    49: {'Abrev.': 'BID', 'Medida': 'Bidón'},
    50: {'Abrev.': 'RAC', 'Medida': 'Ración'},
    51: {'Abrev.': 'Q', 'Medida': 'Quintal'},
    52: {'Abrev.': 'GRT', 'Medida': 'Toneladas de registro bruto'},
    53: {'Abrev.': 'P2', 'Medida': 'Pie cuadrado'},
    54: {'Abrev.': 'PAX', 'Medida': 'Pasajero'},
    55: {'Abrev.': 'PULG', 'Medida': 'Pulgadas'},
    56: {'Abrev.': 'STAY', 'Medida': 'Parqueo barcos en muelle'}
}



from setuptools import setup, find_packages


# Files to be included in the distribution
packages = find_packages()

# Dependences
install_requires = [
    'requests',
]

# Setup configuration
setup(
    name='alanube',
    version='0.0.5',
    author='Wilmer Martinez',
    author_email='info@wilmermartinez.dev',
    description='Library to establish connections with the Alanube API.',
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/wilmerm/alanube-python',
    license='MIT',
    packages=packages,
    install_requires=install_requires,
    keywords=[
        'alanube',
        'api',
        'dgii',
        'ncf',
        'electronic billing',
        'electronic bill',
        'billing',
        'bill',
        'invoice',
        'erp',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)




from ..config import *


# Sólo para Notas de Crédito que no tienen derecho a rebajar ITBIS.
CREDIT_NOTE_INDICATOR_DAYS = 30
CREDIT_NOTE_INDICATOR_EXPIRED = 1
"""Valor si fecha de emisión del e-CF afectado es > `CREDIT_NOTE_INDICATOR_DAYS`."""
CREDIT_NOTE_INDICATOR_NOT_EXPIRED = 0
"""Valor si fecha de emisión del e-CF afectado es <= `CREDIT_NOTE_INDICATOR_DAYS`."""


# Legal Status
REJECTED = 'REJECTED' # El documento fue rechazado
ACCEPTED = 'ACCEPTED' # El documento fue aceptado
ACCEPTED_WITH_OBSERVATIONS = 'ACCEPTED_WITH_OBSERVATIONS' # El documento fue aceptado parcialmente
IN_PROCESS = 'IN_PROCESS' # El documento se encuentra en proceso

LEGAL_STATUS_LIST = [
    REJECTED,
    ACCEPTED,
    ACCEPTED_WITH_OBSERVATIONS,
    IN_PROCESS,
]


FIT_TYPE_DISCOUNT = 'D'
FIT_TYPE_SURCHARGE = 'R'

FIT_TYPES = [
    FIT_TYPE_DISCOUNT,
    FIT_TYPE_SURCHARGE,
]


# Identifica si el ítem corresponde a Bien o Servicio.
ITEM_GOOD_INDICATOR = 1
ITEM_SERVICE_INDICATOR = 2

ITEM_GOOD_SERVICE_INDICATORS = [
    ITEM_GOOD_INDICATOR,
    ITEM_SERVICE_INDICATOR,
]


# Indica si el ítem es exento, si es gravado, o No facturable. Indicará las distintas tasas:
ITEM_BILLING_INDICATOR_0 = 0
"""No Facturable"""
ITEM_BILLING_INDICATOR_1 = 1
"""ITBIS 1 ítem gravado a ITBIS tasa1 (18%)."""
ITEM_BILLING_INDICATOR_2 = 2
"""ITBIS 1 ítem gravado a ITBIS tasa1 (16%)."""
ITEM_BILLING_INDICATOR_3 = 3
"""ITBIS 1 ítem gravado a ITBIS tasa1 (0%)."""
ITEM_BILLING_INDICATOR_4 = 4
"""Excento (E)"""

ITEM_BILLING_INDICATORS = {
    ITEM_BILLING_INDICATOR_0: 0,
    ITEM_BILLING_INDICATOR_1: 18,
    ITEM_BILLING_INDICATOR_2: 16,
    ITEM_BILLING_INDICATOR_3: 0,
    ITEM_BILLING_INDICATOR_4: 0,
}


# Tipos de documentos (type)
# Tipos de documentos enviados en el webhook de documento finalizado.
fiscalInvoice = 'fiscalInvoice'	# Factura de crédito fiscal (31)
invoice	= 'invoice' # Factura de consumo (32)
debitNote = 'debitNote' # Notas débito (33)
creditNote = 'creditNote' # Notas crédito (34)
specialRegime = 'specialRegime' # Factura para regímenes especiales (44)
gubernamental = 'gubernamental' # Facturas Gubernamentales (45)
exportSupport = 'exportSupport' # Facturas de Exportación (46)
purchase = 'purchase' # Compras (41)
minorExpense = 'minorExpense' # Gastos menores (43)
paymentAbroadSupport = 'paymentAbroadSupport' # Comprobante para Pagos al Exterior (47)

INVOICE_CODE_TYPES = {
    31: fiscalInvoice,
    32: invoice,
    33: debitNote,
    34: creditNote,
    44: specialRegime,
    45: gubernamental,
    46: exportSupport,
    41: purchase,
    43: minorExpense,
    47: paymentAbroadSupport
}


INCOME_TYPE_01 = 1 # Ingresos por operaciones (no financieros)
INCOME_TYPE_02 = 2 # Ingresos financieros
INCOME_TYPE_03 = 3 # Ingresos extraordinarios
INCOME_TYPE_04 = 4 # Ingresos por arrendamientos
INCOME_TYPE_05 = 5 # Ingresos por venta de activo depreciable
INCOME_TYPE_06 = 6 # Otros ingresos

INCOME_TYPES = [
    INCOME_TYPE_01,
    INCOME_TYPE_02,
    INCOME_TYPE_03,
    INCOME_TYPE_04,
    INCOME_TYPE_05,
    INCOME_TYPE_06,
]


MODIFICATION_CODE_1 = 1
"""1) Anulación total"""
MODIFICATION_CODE_2 = 2
"""2) Corrige texto del comprobante fiscal modificado"""
MODIFICATION_CODE_3 = 3
"""3) Corrige montos del NCF modificado"""
MODIFICATION_CODE_4 = 4
"""4) Reemplazo NCF emitido en contingencia"""
MODIFICATION_CODE_5 = 5
"""5) Referenciar Factura de Consumo Electrónica."""

MODIFICATION_CODES = [
    MODIFICATION_CODE_1,
    MODIFICATION_CODE_2,
    MODIFICATION_CODE_3,
    MODIFICATION_CODE_4,
    MODIFICATION_CODE_5,
]
"""
Código utilizado para indicar si el e-CF del comprobante fiscal modificado es con la finalidad de:
1) Anulación total
2) Corrige texto del comprobante fiscal modificado
3) Corrige montos del NCF modificado
4) Reemplazo NCF emitido en contingencia
5) Referenciar Factura de Consumo Electrónica.
"""


NCF_B = 'B'
NCF_B_01 = '01'
NCF_B_02 = '02'
NCF_B_03 = '03'
NCF_B_04 = '04'
NCF_B_11 = '11'
NCF_B_12 = '12'
NCF_B_13 = '13'
NCF_B_14 = '14'
NCF_B_15 = '15'
NCF_B_16 = '16'
NCF_B_17 = '17'

NCF_E = 'E'
"""Serie de Comprobantes Fiscales Electrónicos (e-CF)"""
NCF_E_31 = '31'
"""Factura de Crédito Fiscal Electrónico (Tipo 31)"""
NCF_E_32 = '32'
"""Factura de Consumo Electrónica (Tipo 32)"""
NCF_E_33 = '33'
"""Nota de Débito Electrónica (Tipo 33)"""
NCF_E_34 = '34'
"""Nota de Crédito Electrónica (Tipo 34)"""
NCF_E_41 = '41'
"""Compras Electrónico (Tipo 41)"""
NCF_E_43 = '43'
"""Gastos Menores Electrónico (Tipo 43)"""
NCF_E_44 = '44'
"""Regímenes Especiales Electrónico (Tipo 44)"""
NCF_E_45 = '45'
"""Gubernamental Electrónico (Tipo 45)"""
NCF_E_46 = '46'
"""Exportación Electrónico (Tipo 46)"""
NCF_E_47 = '47'
"""Pagos Exterior Electrónico (Tipo 47)"""

NCF_SERIES = {
    NCF_B: {
        'length': 11,
        'codes': [
            NCF_B_01,
            NCF_B_02,
            NCF_B_03,
            NCF_B_04,
            NCF_B_11,
            NCF_B_12,
            NCF_B_13,
            NCF_B_14,
            NCF_B_15,
            NCF_B_16,
            NCF_B_17,
        ],
    },
    NCF_E: {
        'length': 13,
        'codes': [
            NCF_E_31,
            NCF_E_32,
            NCF_E_33,
            NCF_E_34,
            NCF_E_41,
            NCF_E_43,
            NCF_E_44,
            NCF_E_45,
            NCF_E_46,
            NCF_E_47,
        ]
    }
}


NUMBER_DEFAULT_MAX_DIGIT = 18
NUMBER_DEFAULT_DECIMAL_PLACES = 2
NUMBER_DEFAULT_MAX_VALUE = 9999999999999999.99


PAYMENT_TYPE_CASH = 1
PAYMENT_TYPE_CREDIT = 2
PAYMENT_TYPE_FREE = 3

PAYMENT_TYPES = [
    PAYMENT_TYPE_CASH,
    PAYMENT_TYPE_CREDIT,
    PAYMENT_TYPE_FREE,
]

PAYMENT_METHOD_CASH = 1
PAYMENT_METHOD_TRANSFER = 2
PAYMENT_METHOD_CREDIT_CARD = 3
PAYMENT_METHOD_CREDIT = 4
PAYMENT_METHOD_BONUS = 5
PAYMENT_METHOD_SWAP = 6
PAYMENT_METHOD_CREDIT_NOTE = 7
PAYMENT_METHOD_OTHERS = 8

PAYMENT_METHODS = [
    PAYMENT_METHOD_CASH,
    PAYMENT_METHOD_TRANSFER,
    PAYMENT_METHOD_CREDIT_CARD,
    PAYMENT_METHOD_CREDIT,
    PAYMENT_METHOD_BONUS,
    PAYMENT_METHOD_SWAP,
    PAYMENT_METHOD_CREDIT_NOTE,
    PAYMENT_METHOD_OTHERS,
]

PAYMENT_ACCOUNT_TYPE_CT = 'CT'
"""Cuenta Corriente"""
PAYMENT_ACCOUNT_TYPE_AH = 'AH'
"""Cuenta de Ahorros"""
PAYMENT_ACCOUNT_TYPE_OT = 'OT'
"""Otra"""

PAYMENT_ACCOUNT_TYPES = [
    PAYMENT_ACCOUNT_TYPE_CT,
    PAYMENT_ACCOUNT_TYPE_AH,
    PAYMENT_ACCOUNT_TYPE_OT,
]


PROVINCES_AND_MUNICIPALITIES = {
    '010000': {
        'name': 'DISTRITO NACIONAL',
        'municipalities': {
            '010100': 'MUNICIPIO SANTO DOMINGO DE GUZMÁN',
            '010101': 'SANTO DOMINGO DE GUZMÁN (D. M.).',
        }
    },
    '020000': {
        'name': 'PROVINCIA AZUA',
        'municipalities': {
            '020100': 'MUNICIPIO AZUA',
            '020101': 'AZUA (D. M.).',
            '020102': 'BARRO ARRIBA (D. M.).',
            '020103': 'LAS BARÍAS-LA ESTANCIA (D. M.).',
            '020104': 'LOS JOVILLOS (D. M.).',
            '020105': 'PUERTO VIEJO (D. M.).',
            '020106': 'BARRERAS (D. M.).',
            '020107': 'DOÑA EMMA BALAGUER VIUDA VALLEJO (D. M.).',
            '020108': 'CLAVELLINA (D. M.).',
            '020109': 'LAS LOMAS (D. M.).',
            '020200': 'MUNICIPIO LAS CHARCAS',
            '020201': 'LAS CHARCAS (D. M.).',
            '020202': 'PALMAR DE OCOA (D. M.).',
            '020300': 'MUNICIPIO LAS YAYAS DE VIAJAMA',
            '020301': 'LAS YAYAS DE VIAJAMA (D. M.).',
            '020302': 'VILLARPANDO (D. M.).',
            '020303': 'HATO NUEVO CORTÉS (D. M.).',
            '020400': 'MUNICIPIO PADRE LAS CASAS',
            '020401': 'PADRE LAS CASAS (D. M.).',
            '020402': 'LAS LAGUNAS (D. M.).',
            '020403': 'LA SIEMBRA (D. M.).',
            '020404': 'MONTE BONITO (D. M.).',
            '020405': 'LOS FRÍOS (D. M.).',
            '020500': 'MUNICIPIO PERALTA',
            '020501': 'PERALTA (D. M.).',
            '020600': 'MUNICIPIO SABANA YEGUA',
            '020601': 'SABANA YEGUA (D. M.).',
            '020602': 'PROYECTO 4 (D. M.).',
            '020603': 'GANADERO (D. M.).',
            '020604': 'PROYECTO 2-C (D. M.).',
            '020700': 'MUNICIPIO PUEBLO VIEJO',
            '020701': 'PUEBLO VIEJO (D. M.).',
            '020702': 'EL ROSARIO (D. M.).',
            '020800': 'MUNICIPIO TÁBARA ARRIBA',
            '020801': 'TÁBARA ARRIBA (D. M.).',
            '020802': 'TÁBARA ABAJO (D. M.).',
            '020803': 'AMIAMA GÓMEZ (D. M.).',
            '020804': 'LOS TOROS (D. M.).',
            '020900': 'MUNICIPIO GUAYABAL',
            '020901': 'GUAYABAL (D. M.).',
            '021000': 'MUNICIPIO ESTEBANÍA',
            '021001': 'ESTEBANÍA (D. M.).',
        },
    },
    '030000': {
        'name': 'PROVINCIA BAHORUCO',
        'municipalities': {
            '030001': 'MUNICIPIO NEIBA',
            '030101': 'NEIBA (D. M.)',
            '030102': 'EL PALMAR (D. M.)',
            '030200': 'MUNICIPIO GALVÁN',
            '030201': 'GALVÁN (D. M.)',
            '030202': 'EL SALADO (D. M.)',
            '030300': 'MUNICIPIO TAMAYO',
            '030301': 'TAMAYO (D. M.)',
            '030302': 'UVILLA (D. M.)',
            '030303': 'SANTANA (D. M.)',
            '030304': 'MONSERRATE (MONTSERRAT) (D. M.)',
            '030305': 'CABEZA DE TORO (D. M.)',
            '030306': 'MENA (D. M.)',
            '030307': 'SANTA BÁRBARA EL 6 (D. M.)',
            '030400': 'MUNICIPIO VILLA JARAGUA',
            '030401': 'VILLA JARAGUA (D. M.)',
            '030500': 'MUNICIPIO LOS RÍOS',
            '030501': 'LOS RÍOS (D. M.)',
            '030502': 'LAS CLAVELLINAS (D. M.)'
        }
    },
    '040000': {
        'name': 'PROVINCIA BARAHONA',
        'municipalities': {
            '040100': 'MUNICIPIO BARAHONA',
            '040101': 'BARAHONA (D. M.)',
            '040102': 'EL CACHÓN (D. M.)',
            '040103': 'LA GUÁZARA (D. M.)',
            '040104': 'VILLA CENTRAL (D. M.)',
            '040200': 'MUNICIPIO CABRAL',
            '040201': 'CABRAL (D. M.)',
            '040300': 'MUNICIPIO ENRIQUILLO',
            '040301': 'ENRIQUILLO (D. M.)',
            '040302': 'ARROYO DULCE (D. M.)',
            '040400': 'MUNICIPIO PARAÍSO',
            '040401': 'PARAÍSO (D. M.)',
            '040402': 'LOS PATOS (D. M.)',
            '040500': 'MUNICIPIO VICENTE NOBLE',
            '040501': 'VICENTE NOBLE (D. M.)',
            '040502': 'CANOA (D. M.)',
            '040503': 'QUITA CORAZA (D. M.)',
            '040504': 'FONDO NEGRO (D. M.)',
            '040600': 'MUNICIPIO EL PEÑÓN',
            '040601': 'EL PEÑÓN (D. M.)',
            '040700': 'MUNICIPIO LA CIÉNAGA',
            '040701': 'LA CIÉNAGA (D. M.)',
            '040702': 'BAHORUCO (D. M.)',
            '040800': 'MUNICIPIO FUNDACIÓN',
            '040801': 'FUNDACIÓN (D. M.)',
            '040802': 'PESCADERÍA (D. M.)',
            '040900': 'MUNICIPIO LAS SALINAS',
            '040901': 'LAS SALINAS (D. M.)',
            '041000': 'MUNICIPIO POLO',
            '041001': 'POLO (D. M.)',
            '041100': 'MUNICIPIO JAQUIMEYES',
            '041101': 'JAQUIMEYES (D. M.)',
            '041102': 'PALO ALTO (D. M.)'
        }
    },
    '050000': {
        'name': 'PROVINCIA DAJABÓN',
        'municipalities': {
            '050100': 'MUNICIPIO DAJABÓN',
            '050101': 'DAJABÓN (D. M.)',
            '050102': 'CAÑONGO (D. M.)',
            '050200': 'MUNICIPIO LOMA DE CABRERA',
            '050201': 'LOMA DE CABRERA (D. M.)',
            '050202': 'CAPOTILLO (D. M.)',
            '050203': 'SANTIAGO DE LA CRUZ (D. M.)',
            '050300': 'MUNICIPIO PARTIDO',
            '050301': 'PARTIDO (D. M.)',
            '050400': 'MUNICIPIO RESTAURACIÓN',
            '050401': 'RESTAURACIÓN (D. M.)',
            '050500': 'MUNICIPIO EL PINO',
            '050501': 'EL PINO (D. M.)',
            '050502': 'MANUEL BUENO (D. M.)'
        }
    },
    '060000': {
        'name': 'PROVINCIA DUARTE',
        'municipalities': {
            '060100': 'MUNICIPIO SAN FRANCISCO DE MACORÍS',
            '060101': 'SAN FRANCISCO DE MACORÍS (D. M.)',
            '060102': 'LA PEÑA (D. M.)',
            '060103': 'CENOVÍ (D. M.)',
            '060104': 'JAYA (D. M.)',
            '060105': 'PRESIDENTE DON ANTONIO GUZMÁN FERNÁNDEZ (D. M.)',
            '060200': 'MUNICIPIO ARENOSO',
            '060201': 'ARENOSO (D. M.)',
            '060202': 'LAS COLES (D. M.)',
            '060203': 'EL AGUACATE (D. M.)',
            '060300': 'MUNICIPIO CASTILLO',
            '060301': 'CASTILLO (D. M.)',
            '060400': 'MUNICIPIO PIMENTEL',
            '060401': 'PIMENTEL (D. M.)',
            '060500': 'MUNICIPIO VILLA RIVA',
            '060501': 'VILLA RIVA (D. M.)',
            '060502': 'AGUA SANTA DEL YUNA (D. M.)',
            '060503': 'CRISTO REY DE GUARAGUAO (D. M.)',
            '060504': 'LAS TARANAS (D. M.)',
            '060505': 'BARRAQUITO (D. M.)',
            '060600': 'MUNICIPIO LAS GUÁRANAS',
            '060601': 'LAS GUÁRANAS (D. M.)',
            '060700': 'MUNICIPIO EUGENIO MARÍA DE HOSTOS',
            '060701': 'EUGENIO MARÍA DE HOSTOS (D. M.)',
            '060702': 'SABANA GRANDE (D. M.)'
        }
    },
    '070000': {
        'name': 'PROVINCIA ELÍAS PIÑA',
        'municipalities': {
            '070100': 'MUNICIPIO COMENDADOR',
            '070101': 'COMENDADOR (D. M.)',
            '070102': 'SABANA LARGA (D. M.)',
            '070103': 'GUAYABO (D. M.)',
            '070200': 'MUNICIPIO BÁNICA',
            '070201': 'BÁNICA (D. M.)',
            '070202': 'SABANA CRUZ (D. M.)',
            '070203': 'SABANA HIGÜERO (D. M.)',
            '070300': 'MUNICIPIO EL LLANO',
            '070301': 'EL LLANO (D. M.)',
            '070302': 'GUANITO (D. M.)',
            '070400': 'MUNICIPIO HONDO VALLE',
            '070401': 'HONDO VALLE (D. M.)',
            '070402': 'RANCHO DE LA GUARDIA (D. M.)',
            '070500': 'MUNICIPIO PEDRO SANTANA',
            '070501': 'PEDRO SANTANA (D. M.)',
            '070502': 'RÍO LIMPIO (D. M.)',
            '070600': 'MUNICIPIO JUAN SANTIAGO',
            '070601': 'JUAN SANTIAGO (D. M.)'
        }
    },
    '080000': {
        'name': 'PROVINCIA EL SEIBO',
        'municipalities': {
            '080100': 'MUNICIPIO EL SEIBO',
            '080101': 'EL SEIBO (D. M.)',
            '080102': 'PEDRO SÁNCHEZ (D. M.)',
            '080103': 'SAN FRANCISCO-VICENTILLO (D. M.)',
            '080104': 'SANTA LUCÍA (D. M.)',
            '080200': 'MUNICIPIO MICHES',
            '080201': 'MICHES (D. M.)',
            '080202': 'EL CEDRO (D. M.)',
            '080203': 'LA GINA (D. M.)'
        }
    },
    '090000': {
        'name': 'PROVINCIA ESPAILLAT',
        'municipalities': {
            '090100': 'MUNICIPIO MOCA',
            '090101': 'MOCA (D. M.)',
            '090102': 'JOSÉ CONTRERAS (D. M.)',
            '090103': 'SAN VÍCTOR (D. M.)',
            '090104': 'JUAN LÓPEZ (D. M.)',
            '090105': 'LAS LAGUNAS (D. M.)',
            '090106': 'CANCA LA REYNA (D. M.)',
            '090107': 'EL HIGÜERITO (D. M.)',
            '090108': 'MONTE DE LA JAGUA (D. M.)',
            '090109': 'LA ORTEGA (D. M.)',
            '090200': 'MUNICIPIO CAYETANO GERMOSÉN',
            '090201': 'CAYETANO GERMOSÉN (D. M.)',
            '090300': 'MUNICIPIO GASPAR HERNÁNDEZ',
            '090301': 'GASPAR HERNÁNDEZ (D. M.)',
            '090302': 'JOBA ARRIBA (D. M.)',
            '090303': 'VERAGUA (D. M.)',
            '090304': 'VILLA MAGANTE (D. M.)',
            '090400': 'MUNICIPIO JAMAO AL NORTE',
            '090401': 'JAMAO AL NORTE (D. M.)'
        }
    },
    '100000': {
        'name': 'PROVINCIA INDEPENDENCIA',
        'municipalities': {
            '100100': 'MUNICIPIO JIMANÍ',
            '100101': 'JIMANÍ (D. M.)',
            '100102': 'EL LIMÓN (D. M.)',
            '100103': 'BOCA DE CACHÓN (D. M.)',
            '100200': 'MUNICIPIO DUVERGÉ',
            '100201': 'DUVERGÉ (D. M.)',
            '100202': 'VENGAN A VER (D. M.)',
            '100300': 'MUNICIPIO LA DESCUBIERTA',
            '100301': 'LA DESCUBIERTA (D. M.)',
            '100400': 'MUNICIPIO POSTRER RÍO',
            '100401': 'POSTRER RÍO (D. M.)',
            '100402': 'GUAYABAL (D. M.)',
            '100500': 'MUNICIPIO CRISTÓBAL',
            '100501': 'CRISTÓBAL (D. M.)',
            '100502': 'BATEY 8 (D. M.)',
            '100600': 'MUNICIPIO MELLA',
            '100601': 'MELLA (D. M.)',
            '100602': 'LA COLONIA (D. M.)'
        }
    },
    '110000': {
        'name': 'PROVINCIA LA ALTAGRACIA',
        'municipalities': {
            '110100': 'MUNICIPIO HIGÜEY',
            '110101': 'HIGÜEY (D. M.)',
            '110102': 'LAS LAGUNAS DE NISIBÓN (D. M.)',
            '110103': 'LA OTRA BANDA (D. M.)',
            '110104': 'VERÓN PUNTA CANA (D. M.) (Incluye Bávaro)',
            '110200': 'MUNICIPIO SAN RAFAEL DEL YUMA',
            '110201': 'SAN RAFAEL DEL YUMA (D. M.)',
            '110202': 'BOCA DE YUMA (D. M.)',
            '110203': 'BAYAHÍBE (D. M.)'
        }
    },
    '120000': {
        'name': 'PROVINCIA LA ROMANA',
        'municipalities': {
            '120100': "MUNICIPIO LA ROMANA",
            '120101': "LA ROMANA (D. M.)",
            '120102': "CALETA (D. M.)",
            '120200': "MUNICIPIO GUAYMATE",
            '120201': "GUAYMATE (D. M.)",
            '120300': "MUNICIPIO VILLA HERMOSA",
            '120301': "VILLA HERMOSA (D. M.)",
            '120302': "CUMAYASA (D. M.)"
        }
    },
    '130000': {
        'name': 'PROVINCIA LA VEGA',
        'municipalities': {
            "130100": "MUNICIPIO LA VEGA",
            "130101": "LA VEGA (D. M.)",
            "130102": "RÍO VERDE ARRIBA (D. M.)",
            "130103": "EL RANCHITO (D. M.)",
            "130104": "TAVERAS (D. M.)",
            "130105": "DON JUAN RODRÍGUEZ (D.M.)",
            "130200": "MUNICIPIO CONSTANZA",
            "130201": "CONSTANZA (D. M.)",
            "130202": "TIREO (D. M.)",
            "130203": "LA SABINA (D. M.)",
            "130300": "MUNICIPIO JARABACOA",
            "130301": "JARABACOA (D. M.)",
            "130302": "BUENA VISTA (D. M.)",
            "130303": "MANABAO (D. M.)",
            "130400": "MUNICIPIO JIMA ABAJO",
            "130401": "JIMA ABAJO (D. M.)",
            "130402": "RINCÓN (D. M.)"
        }
    },
    '140000': {
        'name': 'PROVINCIA MARÍA TRINIDAD SÁNCHEZ',
        'municipalities': {
            "140100": "MUNICIPIO NAGUA",
            "140101": "NAGUA (D. M.)",
            "140102": "SAN JOSÉ DE MATANZAS (D. M.)",
            "140103": "LAS GORDAS (D. M.)",
            "140104": "ARROYO AL MEDIO (D. M.)",
            "140200": "MUNICIPIO CABRERA",
            "140201": "CABRERA (D. M.)",
            "140202": "ARROYO SALADO (D. M.)",
            "140203": "LA ENTRADA (D. M.)",
            "140300": "MUNICIPIO EL FACTOR",
            "140301": "EL FACTOR (D. M.)",
            "140302": "EL POZO (D. M.)",
            "140400": "MUNICIPIO RÍO SAN JUAN",
            "140401": "RÍO SAN JUAN (D. M.)"
        }
    },
    '150000': {
        'name': 'PROVINCIA MONTE CRISTI',
        'municipalities': {
            "150100": "MUNICIPIO MONTE CRISTI",
            "150101": "MONTE CRISTI (D. M.)",
            "150200": "MUNICIPIO CASTAÑUELAS",
            "150201": "CASTAÑUELAS (D. M.)",
            "150202": "PALO VERDE (D. M.)",
            "150300": "MUNICIPIO GUAYUBÍN",
            "150301": "GUAYUBÍN (D. M.)",
            "150302": "VILLA ELISA (D. M.)",
            "150303": "HATILLO PALMA (D. M.)",
            "150304": "CANA CHAPETÓN (D. M.)",
            "150400": "MUNICIPIO LAS MATAS DE SANTA CRUZ",
            "150401": "LAS MATAS DE SANTA CRUZ (D. M.)",
            "150500": "MUNICIPIO PEPILLO SALCEDO",
            "150501": "PEPILLO SALCEDO (MANZANILLO)",
            "150502": "SANTA MARÍA (D. M.)",
            "150600": "MUNICIPIO VILLA VÁSQUEZ",
            "150601": "VILLA VÁSQUEZ"
        }
    },
    '160000': {
        'name': 'PROVINCIA PEDERNALES',
        'municipalities': {
            "160100": "MUNICIPIO PEDERNALES",
            "160101": "PEDERNALES",
            "160102": "JOSÉ FRANCISCO PEÑA GÓMEZ (D. M.)",
            "160200": "MUNICIPIO OVIEDO",
            "160201": "OVIEDO",
            "160202": "JUANCHO (D. M.)"
        }
    },
    '170000': {
        'name': 'PROVINCIA PERAVIA',
        'municipalities': {
            "170100": "MUNICIPIO BANÍ",
            "170101": "BANÍ (D. M.)",
            "170102": "MATANZAS (D. M.)",
            "170103": "VILLA FUNDACIÓN (D. M.)",
            "170104": "SABANA BUEY (D. M.)",
            "170105": "PAYA (D. M.)",
            "170106": "VILLA SOMBRERO (D. M.)",
            "170107": "EL CARRETÓN (D. M.)",
            "170108": "CATALINA (D. M.)",
            "170109": "EL LIMONAL (D. M.)",
            "170110": "LAS BARÍAS (D. M.)",
            "170200": "MUNICIPIO NIZAO",
            "170201": "NIZAO",
            "170202": "PIZARRETE (D. M.)",
            "170203": "SANTANA (D. M.)",
            "170300": "MATANZAS",
            "170301": "MATANZAS"
        }
    },
    '180000': {
        'name': 'PROVINCIA PUERTO PLATA',
        'municipalities': {
            "180100": "MUNICIPIO PUERTO PLATA",
            "180101": "PUERTO PLATA (D. M.)",
            "180102": "YÁSICA ARRIBA (D. M.)",
            "180103": "MAIMÓN (D. M.)",
            "180200": "MUNICIPIO ALTAMIRA",
            "180201": "ALTAMIRA",
            "180202": "RÍO GRANDE (D. M.)",
            "180300": "MUNICIPIO GUANANICO",
            "180301": "GUANANICO",
            "180400": "MUNICIPIO IMBERT",
            "180401": "IMBERT",
            "180500": "MUNICIPIO LOS HIDALGOS",
            "180501": "LOS HIDALGOS",
            "180502": "NAVAS (D. M.)",
            "180600": "MUNICIPIO LUPERÓN",
            "180601": "LUPERÓN",
            "180602": "LA ISABELA (D. M.)",
            "180603": "BELLOSO (D. M.)",
            "180604": "EL ESTRECHO DE LUPERÓN OMAR BROSS (D. M.)",
            "180700": "MUNICIPIO SOSÚA",
            "180701": "SOSÚA",
            "180702": "CABARETE (D. M.)",
            "180703": "SABANETA DE YÁSICA (D. M.)",
            "180800": "MUNICIPIO VILLA ISABELA",
            "180801": "VILLA ISABELA",
            "180802": "ESTERO HONDO (D. M.)",
            "180803": "LA JAIBA (D. M.)",
            "180804": "GUALETE (D. M.)",
            "180900": "MUNICIPIO VILLA MONTELLANO",
            "180901": "VILLA MONTELLANO"
        }
    },
    '190000': {
        'name': 'PROVINCIA HERMANAS MIRABAL',
        'municipalities': {
            "190100": "MUNICIPIO SALCEDO",
            "190101": "SALCEDO",
            "190102": "JAMAO AFUERA (D. M.)",
            "190200": "MUNICIPIO TENARES",
            "190201": "TENARES",
            "190202": "BLANCO (D. M.)",
            "190300": "MUNICIPIO VILLA TAPIA",
            "190301": "VILLA TAPIA"
        }
    },
    '200000': {
        'name': 'PROVINCIA SAMANÁ',
        'municipalities': {
            "200100": "MUNICIPIO SAMANÁ",
            "200101": "SAMANÁ",
            "200102": "EL LIMÓN (D. M.)",
            "200103": "ARROYO BARRIL (D. M.)",
            "200104": "LAS GALERAS (D. M.)",
            "200200": "MUNICIPIO SÁNCHEZ",
            "200201": "SÁNCHEZ (D. M.)",
            "200300": "MUNICIPIO LAS TERRENAS",
            "200301": "LAS TERRENAS"
        }
    },
    '210000': {
        'name': 'PROVINCIA SAN CRISTÓBAL',
        'municipalities': {
            "210100": "MUNICIPIO SAN CRISTÓBAL",
            "210101": "SAN CRISTÓBAL (D. M.)",
            "210102": "HATO DAMAS (D. M.)",
            "210103": "HATILLO (D. M.)",
            "210200": "MUNICIPIO SABANA GRANDE DE PALENQUE",
            "210201": "SABANA GRANDE DE PALENQUE (D. M.)",
            "210300": "MUNICIPIO BAJOS DE HAINA",
            "210301": "BAJOS DE HAINA",
            "210302": "EL CARRIL (D. M.)",
            "210303": "QUITA SUEÑO (D. M.)",
            "210400": "MUNICIPIO CAMBITA GARABITOS",
            "210401": "CAMBITA GARABITOS",
            "210402": "CAMBITA EL PUEBLECITO (D. M.)",
            "210500": "MUNICIPIO VILLA ALTAGRACIA",
            "210501": "VILLA ALTAGRACIA",
            "210502": "SAN JOSÉ DEL PUERTO (D. M.)",
            "210503": "MEDINA (D. M.)",
            "210504": "LA CUCHILLA (D. M.)",
            "210600": "MUNICIPIO YAGUATE",
            "210601": "YAGUATE (D. M.)",
            "210602": "DOÑA ANA (D. M.)",
            "210700": "MUNICIPIO SAN GREGORIO DE NIGUA",
            "210701": "SAN GREGORIO DE NIGUA",
            "210800": "MUNICIPIO LOS CACAOS",
            "210801": "LOS CACAOS (D. M.)"
        }
    },
    '220000': {
        'name': 'PROVINCIA SAN JUAN',
        'municipalities': {
            "220100": "MUNICIPIO SAN JUAN",
            "220101": "SAN JUAN",
            "220102": "PEDRO CORTO (D. M.)",
            "220103": "SABANETA (D. M.)",
            "220104": "SABANA ALTA (D. M.)",
            "220105": "EL ROSARIO (D. M.)",
            "220106": "HATO DEL PADRE (D. M.)",
            "220107": "GUANITO (D. M.)",
            "220108": "LA JAGUA (D. M.)",
            "220109": "LAS MAGUANAS-HATO NUEVO (D. M.)",
            "220110": "LAS CHARCAS DE MARÍA NOVA (D. M.)",
            "220111": "LAS ZANJAS (D. M.)",
            "220200": "MUNICIPIO BOHECHÍO",
            "220201": "BOHECHÍO",
            "220202": "ARROYO CANO (D. M.)",
            "220203": "YAQUE (D. M.)",
            "220300": "MUNICIPIO EL CERCADO",
            "220301": "EL CERCADO",
            "220302": "DERRUMBADERO (D. M.)",
            "220303": "BATISTA (D. M.)",
            "220400": "MUNICIPIO JUAN DE HERRERA",
            "220401": "JUAN DE HERRERA",
            "220402": "JÍNOVA (D. M.)",
            "220500": "MUNICIPIO LAS MATAS DE FARFÁN",
            "220501": "LAS MATAS DE FARFÁN",
            "220502": "MATAYAYA (D. M.)",
            "220503": "CARRERA DE YEGUAS (D. M.)",
            "220600": "MUNICIPIO VALLEJUELO",
            "220601": "VALLEJUELO",
            "220602": "JORJILLO (D. M.)"
        }
    },
    '230000': {
        'name': 'PROVINCIA SAN PEDRO DE MACORÍS',
        'municipalities': {
            '230100': 'MUNICIPIO SAN PEDRO DE MACORÍS',
            '230101': 'SAN PEDRO DE MACORÍS',
            '230200': 'MUNICIPIO LOS LLANOS',
            '230201': 'LOS LLANOS',
            '230202': 'EL PUERTO (D. M.)',
            '230203': 'GAUTIER (D. M.)',
            '230300': 'MUNICIPIO RAMÓN SANTANA',
            '230301': 'RAMÓN SANTANA',
            '230400': 'MUNICIPIO CONSUELO',
            '230401': 'CONSUELO',
            '230500': 'MUNICIPIO QUISQUEYA',
            '230501': 'QUISQUEYA',
            '230600': 'MUNICIPIO GUAYACANES',
            '230601': 'GUAYACANES'
        }
    },
    '240000': {
        'name': 'PROVINCIA SÁNCHEZ RAMÍREZ',
        'municipalities': {
            '240100': 'MUNICIPIO COTUÍ',
            '240101': 'COTUÍ',
            '240102': 'QUITA SUEÑO (D. M.)',
            '240103': 'CABALLERO (D. M.)',
            '240104': 'COMEDERO ARRIBA (D. M.)',
            '240105': 'PLATANAL (D. M.)',
            '240106': 'ZAMBRANA ABAJO',
            '240200': 'MUNICIPIO CEVICOS',
            '240201': 'CEVICOS',
            '240202': 'LA CUEVA (D. M.)',
            '240300': 'MUNICIPIO FANTINO',
            '240301': 'FANTINO',
            '240400': 'MUNICIPIO LA MATA',
            '240401': 'LA MATA',
            '240402': 'LA BIJA (D. M.)',
            '240403': 'ANGELINA (D. M.)',
            '240404': 'HERNANDO ALONZO (D. M.)'
        }
    },
    '250000': {
        'name': 'PROVINCIA SANTIAGO',
        'municipalities': {
            '250100': 'MUNICIPIO SANTIAGO',
            '250101': 'SANTIAGO',
            '250102': 'PEDRO GARCÍA (D. M.)',
            '250104': 'BAITOA (D. M.)',
            '250105': 'LA CANELA (D. M.)',
            '250106': 'SAN FRANCISCO DE JACAGUA (D. M.)',
            '250107': 'HATO DEL YAQUE (D. M.)',
            '250200': 'MUNICIPIO BISONÓ',
            '250201': 'VILLA BISONÓ (NAVARRETE) (D. M.)',
            '250300': 'MUNICIPIO JÁNICO',
            '250301': 'JÁNICO',
            '250302': 'JUNCALITO (D. M.)',
            '250303': 'EL CAIMITO (D. M.)',
            '250400': 'MUNICIPIO LICEY AL MEDIO',
            '250401': 'LICEY AL MEDIO',
            '250402': 'LAS PALOMAS (D. M.)',
            '250500': 'MUNICIPIO SAN JOSÉ DE LAS MATAS',
            '250501': 'SAN JOSÉ DE LAS MATAS',
            '250502': 'EL RUBIO (D. M.)',
            '250503': 'LA CUESTA (D. M.)',
            '250504': 'LAS PLACETAS (D. M.)',
            '250600': 'MUNICIPIO TAMBORIL',
            '250601': 'TAMBORIL',
            '250602': 'CANCA LA PIEDRA (D. M.)',
            '250700': 'MUNICIPIO VILLA GONZÁLEZ',
            '250701': 'VILLA GONZÁLEZ',
            '250702': 'PALMAR ARRIBA (D. M.)',
            '250703': 'EL LIMÓN (D. M.)',
            '250800': 'MUNICIPIO PUÑAL',
            '250801': 'PUÑAL',
            '250802': 'GUAYABAL (D. M.)',
            '250803': 'CANABACOA (D. M.)',
            '250900': 'MUNICIPIO SABANA IGLESIA',
            '250901': 'SABANA IGLESIA',
            '251000': 'BAITOA'
        }
    },
    '260000': {
        'name': 'PROVINCIA SANTIAGO RODRÍGUEZ',
        'municipalities': {
            '260100': 'MUNICIPIO SAN IGNACIO DE SABANETA',
            '260101': 'SAN IGNACIO DE SABANETA (D. M.)',
            '260200': 'MUNICIPIO VILLA LOS ALMÁCIGOS',
            '260201': 'VILLA LOS ALMÁCIGOS (D. M.)',
            '260300': 'MUNICIPIO MONCIÓN',
            '260301': 'MONCIÓN (D. M.)'
        }
    },
    '270000': {
        'name': 'PROVINCIA VALVERDE',
        'municipalities': {
            '270100': 'MUNICIPIO MAO',
            '270101': 'MAO (D. M.)',
            '270102': 'AMINA (D. M.)',
            '270103': 'JAIBÓN (PUEBLO NUEVO) (D. M.)',
            '270104': 'GUATAPANAL (D. M.)'
        }
    },
    '280000': {
        'name': 'PROVINCIA MONSEÑOR NOUEL',
        'municipalities': {
            '280100': 'MUNICIPIO BONAO',
            '280101': 'BONAO (D. M.)',
            '280102': 'SABANA DEL PUERTO (D. M.)',
            '280103': 'JUMA BEJUCAL (D. M.)',
            '280104': 'ARROYO TORO - MASIPEDRO (D. M.)'
        }
    },
    '290000': {
        'name': 'PROVINCIA MONTE PLATA',
        'municipalities': {
            '290100': 'MUNICIPIO MONTE PLATA',
            '290101': 'MONTE PLATA (D. M.)',
            '290102': 'DON JUAN (D. M.)',
            '290103': 'CHIRINO (D. M.)',
            '290104': 'BOYÁ (D. M.)'
        }
    },
    '300000': {
        'name': 'PROVINCIA HATO MAYOR',
        'municipalities': {
            '300100': 'MUNICIPIO HATO MAYOR',
            '300101': 'HATO MAYOR (D. M.)',
            '300102': 'YERBA BUENA (D. M.)',
            '300103': 'MATA PALACIO (D. M.)',
            '300104': 'GUAYABO DULCE (D. M.)'
        }
    },
    '310000': {
        'name': 'PROVINCIA SAN JOSÉ DE OCOA',
        'municipalities': {
            '310100': 'MUNICIPIO SAN JOSÉ DE OCOA',
            '310101': 'SAN JOSÉ DE OCOA (D. M.)',
            '310102': 'LA CIÉNAGA (D. M.)',
            '310103': 'NIZAO - LAS AUYAMAS (D. M.)',
            '310104': 'EL PINAR (D. M.)'
        }
    },
    '320000': {
        'name': 'PROVINCIA SANTO DOMINGO',
        'municipalities': {
            '320100': 'MUNICIPIO SANTO DOMINGO ESTE',
            '320101': 'SANTO DOMINGO ESTE (D. M.)',
            '320102': 'SAN LUIS (D. M.)',
            '320200': 'MUNICIPIO SANTO DOMINGO OESTE',
            '320201': 'SANTO DOMINGO OESTE (D. M.)',
            '320300': 'MUNICIPIO SANTO DOMINGO NORTE',
            '320301': 'SANTO DOMINGO NORTE (D. M.)',
            '320302': 'LA VICTORIA (D. M.)',
            '320400': 'MUNICIPIO BOCA CHICA',
            '320401': 'BOCA CHICA (D. M.)',
            '320402': 'LA CALETA (D. M.)',
            '320500': 'MUNICIPIO SAN ANTONIO DE GUERRA',
            '320501': 'SAN ANTONIO DE GUERRA (D. M.)',
            '320502': 'HATO VIEJO (D. M.)',
            '320600': 'MUNICIPIO LOS ALCARRIZOS',
            '320601': 'LOS ALCARRIZOS (D. M.)',
            '320602': 'PALMAREJO-VILLA LINDA (D. M.)',
            '320603': 'PANTOJA (D. M.)',
            '320700': 'MUNICIPIO PEDRO BRAND',
            '320701': 'PEDRO BRAND (D. M.)',
            '320702': 'LA GUÁYIGA (D. M.)',
            '320703': 'LA CUABA (D. M.)'
        }
    }
}